#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
RAG Model Types.

RAG Token and RAG Sequence are outlined in https://arxiv.org/abs/2005.11401

RAG Turn is outlined in https://arxiv.org/abs/2104.07567
"""
from abc import ABC, abstractmethod
import torch
import torch.nn
import torch.nn.functional as F
import torch.cuda
from typing import Any, Dict, List, Optional, Tuple


from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import Batch
from parlai.utils.torch import padded_tensor, FP16_PAD_SIZE

from parlai.agents.rag.modules import RagModel
from parlai.agents.rag.retrievers import Document


def sum_across_turns(
    tensor: torch.Tensor,
    input_turns_cnt: torch.LongTensor,
    discount: Optional[float] = None,
) -> torch.Tensor:
    """
    Sum a tensor across turns.

    :param tensor:
        input tensor, of length equivalent to total turns across all batch items
    :param input turns:
        a batch-length tensor indicating number of turns for each batch item
    :param discount:
        an optional discount factor

    :return tensor:
        return batch-length tensor with turns summed.
    """
    if tensor.dim() > 1:
        new_tensor = tensor.new(input_turns_cnt.size(0), *tensor.shape[1:])
    else:
        new_tensor = tensor.new(input_turns_cnt.size(0))
    offset = 0
    for i, it in enumerate(input_turns_cnt):
        tens = tensor[offset : offset + it]
        if discount is not None:
            tens = tens * (discount ** torch.arange(it).flip(0).to(tensor))
        new_tensor[i] = tens.sum()
        offset += it
    return new_tensor


# Reshape from [batch_size, n_docs, dims] to [batch_size * n_docs, dims]
def _stack_ctxt(tensor: torch.Tensor) -> torch.Tensor:
    return tensor.view(-1, *tensor.shape[2:])


# Reshape from [batch_size * n_docs, dims] to [batch_size, n_docs, dims]
def _unstack_ctxt(tensor: torch.Tensor, n_docs: int) -> torch.Tensor:
    return tensor.view(-1, n_docs, *tensor.shape[1:])


def get_forced_decoder_inputs(
    inputs: torch.LongTensor,
    bsz: int,
    start_idx: int,
    end_idx: int,
    generation_model: str,
    start_param: Optional[torch.nn.Parameter] = None,
) -> torch.LongTensor:
    """
    Return the forced decoder inputs, depending on given parameters.

    These inputs are not formatted for RAG models.

    They merely correspond to the appropriate seq2seq_decoder input.
    """
    if generation_model == 'bart':
        tens = torch.LongTensor([end_idx, start_idx]).to(inputs).detach().expand(bsz, 2)
    elif start_param is not None:
        tens = start_param.detach().expand(bsz, 1).to(inputs)
    else:
        tens = torch.LongTensor([start_idx]).expand(inputs.size(0), 1).to(inputs)
    dec_inputs = torch.cat([tens, inputs], 1)
    return dec_inputs  # type: ignore


class RagModelInterface(ABC):
    """
    Define an interface for the RAG Model Types.
    """

    def __init__(self, opt: Opt, null_idx: int):
        self.opt = opt
        self.generation_model = opt['generation_model']
        self.thorough = opt['thorough']
        self.n_docs = opt['n_docs']
        self.null_idx = null_idx

    ##############################
    # Input/Output Augmentations #
    ##############################
    @abstractmethod
    def get_initial_decoder_input(self, input: torch.LongTensor) -> torch.LongTensor:
        """
        Return the initial input for the decoder during generation.
        """

    @abstractmethod
    def get_initial_forced_decoder_input(
        self,
        bsz: int,
        inputs: torch.LongTensor,
        n_docs: int,
        start_idx: int,
        end_idx: int,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.LongTensor:
        """
        Return the initial input to the decoder during training.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode
        :param n_docs:
            number of docs per input
        :param start_idx:
            start token idx
        :param end_idx:
            end token idx
        :param input_turns_cnt:
            an optional tensor containing the number of turns of each corresponding context.

        :return initial_input:
            initial input for the decoder.
        """

    @abstractmethod
    def set_input_turn_cnt_vec(
        self, observation: Message, model: RagModel, query_str: str
    ) -> Message:
        """
        Optionally set the input turn vec.
        """

    ############################
    # Generation Augmentations #
    ############################
    @abstractmethod
    def augment_batch_for_generation(self, batch: Batch, model: RagModel) -> Batch:
        """
        Optionally augment the batch for generation.

        :param batch:
            batch to augment
        :param model:
            model to possibly help with augmenting

        :return batch:
            return batch with appropriate augmentations.
        """

    @abstractmethod
    def rerank_beams(
        self,
        model: RagModel,
        batch: Batch,
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
    ) -> List[List[Tuple[torch.LongTensor, torch.Tensor]]]:
        """
        Optionally rerank beams.
        """

    @abstractmethod
    def reorder_encoder_states(
        self, encoder_states: Tuple[torch.Tensor, ...], indices: torch.LongTensor
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        List[List[Document]],
        torch.Tensor,
    ]:
        """
        Reorder the encoder states, for beam search.

        See ``TorchGeneratorModel.reorder_encoder_states`` for a description.
        """

    ###########################################
    # Loss Computation/Output Marginalization #
    ###########################################
    @abstractmethod
    def compute_loss(
        self,
        criterion: torch.nn.Module,
        scores: torch.Tensor,
        preds: torch.LongTensor,
        enc_state: Tuple[Any],
        label_vec: torch.LongTensor,
    ) -> Tuple[torch.Tensor, List[int], torch.Tensor, torch.Tensor]:
        """
        Compute Loss.

        :param criterion:
            torch criterion module
        :param scores:
            model scores
        :param preds:
            model "predicions" of tokens
        :param enc_state:
            encoder states
        :param label_vec:
            target tokens

        :return (loss, metric_loss, correct_tokens, target_tokens):
            loss: the loss through which we backpropagate
            metric_loss: loss we use for metrics
            correct_tokens: correct predictions from the model
            target_tokens: the ground truth tokens.
        """

    @abstractmethod
    def marginalize(
        self,
        out_probs: torch.Tensor,
        doc_probs: torch.Tensor,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        Marginalize over doc scores.

        :param out_probs:
            tensor of dimension [bsz, n_docs, seqlen, voc_dim]
        :param doc_probs:
            tensor of dimension [bsz, n_docs]
        :param input_turns_cnt:
            optional tensor indicating # of input turns per item.

        :return output:
            return decoder output with docs marginalized.
        """


class RagSequence(RagModelInterface):
    """
    Provides an interface for RAG-Sequence.

    RAG Sequence considers documents independently during training and inference.

    At train time, the document probability is incorporated via addition to the first token
    scores; this allows for backprop to the retriever.

    At inference time, document scores are incorporated only during final beam selection.
    """

    def get_initial_decoder_input(self, input: torch.LongTensor) -> torch.LongTensor:
        """
        Don't repeat decoder input for rag sequence.
        """
        return input

    def get_initial_forced_decoder_input(
        self,
        bsz: int,
        inputs: torch.LongTensor,
        n_docs: int,
        start_idx: int,
        end_idx: int,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.LongTensor:
        """
        Return the initial input to the decoder during training.

        Repeat each input n_docs times.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode
        :param n_docs:
            number of docs per input
        :param start_idx:
            start token idx
        :param end_idx:
            end token idx
        :param input_turns_cnt:
            an optional tensor containing the number of turns of each corresponding context.

        :return initial_input:
            initial input for the decoder.
        """
        inputs = get_forced_decoder_inputs(
            inputs, bsz, start_idx, end_idx, self.generation_model
        )
        inputs = inputs.repeat(1, n_docs).reshape(-1, inputs.size(1))  # type: ignore
        return inputs

    def get_generation_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        Optional[torch.LongTensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
    ]:
        """
        For RAG Sequence, we retrieve prior to generation.
        """
        assert batch.text_vec is not None
        return (batch.text_vec, None, None, None)

    def reorder_encoder_states(
        self, encoder_states: Tuple[torch.Tensor, ...], indices: torch.LongTensor
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        List[List[Document]],
        torch.Tensor,
    ]:
        """
        Reorder the encoder states.

        For RAG Sequence, only reorder enc_out and mask.
        """
        enc, mask, *_ = encoder_states
        enc = torch.index_select(enc, 0, indices)
        mask = torch.index_select(mask, 0, indices)
        return enc, mask, None, None, None  # type: ignore

    def get_ctxt_index(self, batch: Batch, batch_idx: int) -> int:
        """
        Map the batch_idx back to the appropriate batch item during generation.

        :param batch:
            batch being considered
        :param batch_idx:
            idx into batch.text_vec

        :return mapped_idx:
            return the appropriate batch index
        """
        n_docs = batch.doc_log_probs.size(1)  # type: ignore
        return batch_idx // n_docs

    def rerank_beams(
        self,
        model: RagModel,
        batch: Batch,
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
    ) -> List[List[List[torch.LongTensor]]]:
        """
        Rerank beams in RAG-Sequence, accounting for document probabilities as well.

        Fast decoding combines probabilities of equivalently generated beams across
        different documents.

        Thorough decoding rescores the beam outputs with another forward pass.

        Iteration for fast decoding is as follows:
            for each example x_i:
                for each document d_i:
                    marginalize probability over
                    gen. beams. [b_i, ..., b_n] for d_i

        Iteration for thorough decoding is as follows:
            for each example x_i:
                obtain unique hypotheses
                rescore with model, xs=text_vec, ys=hyps
                sort by lowest loss (highest score)

        Thorough decoding impl. verified via OSS impl:
        https://github.com/huggingface/transformers/blob/master/src/transformers/models/rag/modeling_rag.py#L954

        :param batch:
            current batch
        :param n_best_beam_preds_scores:
            bsz-length list of Tuples of predictions and scores

        :return List((pred, score)):
            return a re-ranked version of the n_best_beam_preds_scores
        """
        new_n_best: List[List[List[torch.LongTensor]]] = []
        doc_log_probs = batch.doc_log_probs  # type: ignore
        bsz, n_docs = doc_log_probs.shape
        for i in range(bsz):
            if self.thorough:
                hyps = [
                    hyp[0]
                    for h in n_best_beam_preds_scores[i * n_docs : (i + 1) * n_docs]
                    for hyp in h
                ]
                sorted_by_score = self.thorough_generation(
                    hyps,
                    batch.src_text_vec[i : i + 1],
                    self.null_idx,
                    model,  # type: ignore
                )
            else:
                # mapping from tokens to (tokens, score)
                sorted_by_score = self.fast_generation(
                    list(range(i * n_docs, (i + 1) * n_docs)),
                    n_best_beam_preds_scores,
                    doc_log_probs[i],
                    n_docs,
                )
            new_n_best.append(sorted_by_score)  # type: ignore

        return new_n_best

    def augment_batch_for_generation(self, batch: Batch, model: RagModel) -> Batch:
        """
        Augment batch for generation.

        For RAG Sequence, we retrieve prior to generation, as we do not consider the
        document probabilities until after generating all of the beams.

        :param batch:
            batch to augment
        :param model:
            model to possibly help with augmenting

        :return batch:
            return batch with text vec swapped out.
        """
        (expanded_input, _, doc_scores) = model.retrieve_and_concat(
            batch.text_vec,
            batch.text_vec.ne(self.null_idx).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
        )
        doc_log_probs = F.log_softmax(doc_scores, dim=1)
        batch.src_text_vec = batch.text_vec
        batch.text_vec = expanded_input
        batch.doc_log_probs = doc_log_probs
        batch.batchsize = batch.text_vec.size(0)

        return batch

    @classmethod
    def fast_generation(
        cls,
        doc_indices: List[int],
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
        doc_log_probs: torch.Tensor,
        n_docs: int,
    ):
        """
        Apply RAG Sequence fast decoding, for a single batch item.

        :param doc_indices:
            indices into beam preds scores
        :param n_best_beam_preds_scores:
            bsz-length list of Tuples of predictions and scores
        :param doc_log_probs:
            probabilities for each document
        :param n_docs:
            number of docs per example

        :return sorted_hyps:
            return list of (hyp, score) tuples, sorted by their score.
        """
        marginalized_hypos: Dict[str, List[torch.Tensor]] = {}
        for doc_idx in doc_indices:
            doc_hypos = n_best_beam_preds_scores[doc_idx]
            doc_score = doc_log_probs[doc_idx % n_docs]
            for hypo, hypo_score in doc_hypos:
                score = hypo_score + doc_score
                hypo_tokens = str(hypo.tolist())
                if hypo_tokens in marginalized_hypos:
                    marginalised_hypo = marginalized_hypos[hypo_tokens]
                    marginalised_hypo[1] = torch.log(
                        marginalised_hypo[1].exp() + score.exp()
                    )
                else:
                    marginalized_hypos[hypo_tokens] = [hypo, score]
        sorted_by_score = sorted(marginalized_hypos.values(), key=lambda h: -h[1])
        return sorted_by_score

    @classmethod
    def thorough_generation(
        cls,
        hyps: List[torch.LongTensor],
        new_input: torch.LongTensor,
        null_idx: int,
        model: RagModel,
    ) -> List[Tuple[torch.LongTensor, torch.Tensor]]:
        """
        Apply RAG-sequence thorough generation for a single batch item.

        Recomputes model scores with given hypotheses, sorts accordingly.

        :param hyps:
            list of candidate hypotheses
        :param new_input:
            input for the model

        :return sorted_hyps:
            return list of (hyp, score) tuples, sorted by their score.
        """
        # deduplicate, exclude BOS Token
        hyps = list({str(h.tolist()): h[1:] for h in hyps}.values())  # type: ignore
        new_input = new_input.repeat(len(hyps), 1)  # type: ignore
        new_ys, _ = padded_tensor(
            hyps, fp16friendly=new_input.size(1) % FP16_PAD_SIZE == 0, pad_idx=null_idx
        )
        new_ys = new_ys.to(new_input.device)
        scores, *_ = model.seq2seq_forward_pass(new_input, new_ys)
        loss = cls._rag_sequence_loss(
            new_ys.unsqueeze(1).unsqueeze(-1), scores.unsqueeze(1), null_idx
        )  # type: ignore
        sorted_by_score = [
            (hyps[idx], loss[idx]) for idx in loss.sort()[-1]
        ]  # sort ascending
        return sorted_by_score

    @classmethod
    def _rag_sequence_loss(
        cls, target: torch.LongTensor, scores: torch.Tensor, null_idx: int
    ) -> torch.Tensor:
        """
        RAG Sequence loss.

        :param target:
            target tokens
        :param scores:
            model log probs
        :param null_idx:
            padding index to ignore for loss computation

        :return loss:
            return NLL Loss
        """
        ll = scores.gather(dim=-1, index=target)
        pad_mask = target.eq(null_idx)
        if pad_mask.any():
            ll.masked_fill_(pad_mask, 0.0)
        ll = ll.squeeze(-1)
        ll = ll.sum(2)  # sum over tokens
        ll = ll.logsumexp(1)  # sum over docs
        nll_loss = -ll
        loss = nll_loss

        return loss

    def set_input_turn_cnt_vec(
        self, observation: Message, model: RagModel, query_str: str
    ) -> Message:
        """
        Ignore for RAG Sequence / Token.
        """
        return observation

    def compute_loss(
        self,
        criterion: torch.nn.Module,
        scores: torch.Tensor,
        preds: torch.LongTensor,
        enc_state: Tuple[Any],
        label_vec: torch.LongTensor,
    ) -> Tuple[torch.Tensor, List[int], torch.Tensor, torch.Tensor]:
        """
        Compute RAG Sequence Loss.

        :param criterion:
            torch criterion module
        :param scores:
            model scores
        :param preds:
            model "predicions" of tokens
        :param enc_state:
            encoder states
        :param label_vec:
            target tokens

        :return (loss, metric_loss, correct_tokens, target_tokens):
            loss: the loss through which we backpropagate
            metric_loss: loss we use for metrics
            correct_tokens: correct predictions from the model
            target_tokens: the ground truth tokens.
        """
        if scores.size(2) != label_vec.size(1):
            assert self.generation_model == 'bart'
            # ignore start
            scores = scores[:, :, 1:, :]
            preds = preds[:, 1:]  # type: ignore

        # compute rag sequence loss
        seq_preds = scores.max(-1)[-1]
        bsz = scores.size(0)
        n_docs = scores.size(1)
        target = label_vec.unsqueeze(1).unsqueeze(-1).repeat(1, n_docs, 1, 1)
        loss = self._rag_sequence_loss(target, scores, self.null_idx)  # type: ignore

        # compute relevant metric counters
        metric_loss = loss.tolist()
        notnull = label_vec.ne(self.null_idx)
        target_tokens = notnull.long().sum(dim=-1)

        shape = (bsz, n_docs, -1)
        notnull_seq = target.ne(self.null_idx)
        correct = (
            (target.view(*shape) == seq_preds.view(*shape)) * notnull_seq.view(*shape)
        ).sum(dim=-1)
        correct_mean = correct.float().mean(dim=1)

        loss = loss.sum()
        loss /= target_tokens.sum()  # average loss per token
        return loss, metric_loss, correct_mean, target_tokens

    def marginalize(
        self,
        out_probs: torch.Tensor,
        doc_probs: torch.Tensor,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        RAG Sequence marginalization over doc scores.

        RAG Sequence simply adds the doc scores to the first token probabilities.

        :param out_probs:
            tensor of dimension [bsz, n_docs, seqlen, voc_dim]
        :param doc_probs:
            tensor of dimension [bsz, n_docs]
        :param input_turns_cnt:
            optional tensor indicating # of input turns per item.

        :return output:
            return decoder output with docs marginalized.
        """
        doc_probs = doc_probs.unsqueeze(-1).unsqueeze(-1)
        first_token_scores = out_probs[:, :, :1, :]
        if self.generation_model == 'bart':
            # bypass first token in BART
            second_token_scores = out_probs[:, :, 1:2, :]
            remainder = out_probs[:, :, 2:, :]
            output = torch.cat(
                [first_token_scores, second_token_scores + doc_probs, remainder], dim=2
            )
        else:
            remainder = out_probs[:, :, 1:, :]
            output = torch.cat([first_token_scores + doc_probs, remainder], dim=2)
        return output


class RagToken(RagModelInterface):
    """
    Provides an interface for RAG-Token.

    RAG Token considers documents jointly during training and inference; output
    distributions are computed via summing across latent document probabilities.
    """

    def get_initial_decoder_input(self, input: torch.LongTensor) -> torch.LongTensor:
        """
        Repeat the decoder input accordingly.
        """
        return input.repeat(1, self.n_docs).reshape(-1, input.size(1))  # type: ignore

    def get_initial_forced_decoder_input(
        self,
        bsz: int,
        inputs: torch.LongTensor,
        n_docs: int,
        start_idx: int,
        end_idx: int,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.LongTensor:
        """
        Return the initial input to the decoder during training.

        Repeat inputs n_docs times.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode
        :param n_docs:
            number of docs per input
        :param start_idx:
            start token idx
        :param end_idx:
            end token idx
        :param input_turns_cnt:
            an optional tensor containing the number of turns of each corresponding context.

        :return initial_input:
            initial input for the decoder.
        """
        inputs = get_forced_decoder_inputs(
            inputs, bsz, start_idx, end_idx, self.generation_model
        )
        inputs = inputs.repeat(1, n_docs).reshape(-1, inputs.size(1))  # type: ignore
        return inputs

    def get_next_decoder_input(
        self,
        prev_input: torch.LongTensor,
        selection: torch.LongTensor,
        incr_state_inds: torch.LongTensor,
    ) -> torch.LongTensor:
        """
        Repeat decoder input accordingly for RAG Token.

        RAG Token marginalizes over all the documents each turn; this means that during generation,
        although the input will be of length [n_docs*bsz], the output will be of length [bsz].

        Thus, we need to make sure that we repeat the beam selection n_docs times and continue to
        marginalize over documents accordingly.

        :param prev_input:
            previous [bsz*n_docs, seqlen] input to the decoder.
        :param selection:
            [bsz]-length beam selection ids for the decoder
        :param incr_state_inds:
            beam indices that are continuing in next generation.

        :return dec_input:
            return the decoder input with appropriate repeated selections.
        """
        prev_input = _unstack_ctxt(prev_input, self.n_docs).index_select(
            0, incr_state_inds
        )  # type: ignore
        dec_input = torch.cat(
            [prev_input, selection.repeat_interleave(self.n_docs, 1).unsqueeze(-1)],
            dim=-1,
        )
        dec_input = _stack_ctxt(dec_input)
        return dec_input  # type: ignore

    def reorder_encoder_states(
        self, encoder_states: Tuple[torch.Tensor, ...], indices: torch.LongTensor
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        List[List[Document]],
        torch.Tensor,
    ]:
        """
        Reorder all RAG encoder states.
        """
        enc, mask, input_turns_cnt, docs, doc_probs = encoder_states
        n_docs = doc_probs.shape[1]
        enc = _stack_ctxt(_unstack_ctxt(enc, n_docs).index_select(0, indices))
        mask = _stack_ctxt(_unstack_ctxt(mask, n_docs).index_select(0, indices))
        doc_probs = doc_probs.index_select(0, indices)

        return enc, mask, input_turns_cnt, docs, doc_probs  # type: ignore

    def rerank_beams(
        self,
        model: RagModel,
        batch: Batch,
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
    ) -> List[List[Tuple[torch.LongTensor, torch.Tensor]]]:
        """
        We don't re-rank beams for RAG Token.
        """
        return n_best_beam_preds_scores

    def augment_batch_for_generation(self, batch: Batch, model: RagModel) -> Batch:
        """
        No augmentation for RAG Token.
        """
        return batch

    def set_input_turn_cnt_vec(
        self, observation: Message, model: RagModel, query_str: str
    ) -> Message:
        """
        Ignore for RAG Sequence / Token.
        """
        return observation

    def compute_loss(
        self,
        criterion: torch.nn.Module,
        scores: torch.Tensor,
        preds: torch.LongTensor,
        enc_state: Tuple[Any],
        label_vec: torch.LongTensor,
    ) -> Tuple[torch.Tensor, List[int], torch.Tensor, torch.Tensor]:
        """
        Compute RAG Token Loss.

        This is a simple NLL Loss.

        :param criterion:
            presumably the NLL criterion.
        :param scores:
            model scores
        :param preds:
            model "predicions" of tokens
        :param enc_state:
            encoder states
        :param label_vec:
            target tokens

        :return (loss, metric_loss, correct_tokens, target_tokens):
            loss: the loss through which we backpropagate
            metric_loss: loss we use for metrics
            correct_tokens: correct predictions from the model
            target_tokens: the ground truth tokens.
        """
        if scores.size(1) != label_vec.size(1):
            assert self.generation_model == 'bart'
            # ignore start
            scores = scores[:, 1:, :]
            preds = preds[:, 1:]  # type: ignore

        # compute loss
        score_view = scores.reshape(-1, scores.size(-1))
        loss = criterion(score_view, label_vec.view(-1))
        loss = loss.view(scores.shape[:-1]).sum(dim=1)

        # calculate metric counters
        metric_loss = loss.tolist()
        notnull = label_vec.ne(self.null_idx)
        target_tokens = notnull.long().sum(dim=-1)
        correct = ((label_vec == preds) * notnull).sum(dim=-1)

        loss = loss.sum()
        loss /= target_tokens.sum()  # average loss per token
        return loss, metric_loss, correct, target_tokens

    def marginalize(
        self,
        out_probs: torch.Tensor,
        doc_probs: torch.Tensor,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        RAG Token marginalization.

        RAG Token sums token probabilities across documents, accounting for each document
        in its final output distribution.

        :param out_probs:
            tensor of dimension [bsz, n_docs, seqlen, voc_dim]
        :param doc_probs:
            tensor of dimension [bsz, n_docs]
        :param input_turns_cnt:
            optional tensor indicating # of input turns per item.

        :return tensor:
            return decoder output with docs marginalized.
        """
        log_prob_sum = out_probs + doc_probs.unsqueeze(-1).unsqueeze(-1)
        output = torch.logsumexp(log_prob_sum, dim=1)  # sum across documents
        return output


class RagTurn(RagModelInterface):
    """
    Provides an interface for RAG-Turn.

    RAG Turn performs retrieval over each turn of dialogue, and marginalizes
    over n_turns context/document pairs for each batch item.

    RAG Turn Doc-Then-Turn marginalizes over documents, then over turns.

    RAG Turn Doc Only considers each turn's documents independently, and thus
    resembles RAG Token on a doc level, and RAG Sequence on a turn level.
    """

    def __init__(self, opt: Opt, null_idx: int):
        super().__init__(opt, null_idx)
        self.turn_marginalize = opt['rag_turn_marginalize']
        self.n_turns = opt['rag_turn_n_turns']
        assert 0 < opt['rag_turn_discount_factor'] <= 1.0
        self.discount_factor = opt['rag_turn_discount_factor']

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        rag_turn_group = parser.add_argument_group('RAG-Turn Args')
        rag_turn_group.add_argument(
            '--rag-turn-n-turns',
            type=int,
            default=2,
            help='how many turns to split up retrieval into. '
            'The most recent text is split by delimiter; all turns after (n-1)th turn '
            'are combined.',
        )
        rag_turn_group.add_argument(
            '--rag-turn-marginalize',
            type=str,
            default='doc_then_turn',
            choices=['doc_only', 'doc_then_turn'],
            help='how to marginalize rag-turn. ',
        )
        rag_turn_group.add_argument(
            '--rag-turn-discount-factor',
            type=float,
            default=1.0,
            help='discount factor for turns beyond most recent one. We employ exponential discounting. '
            'Only considered if 0 < factor < 1.0. ',
        )
        return parser

    def get_initial_decoder_input(self, input: torch.LongTensor) -> torch.LongTensor:
        """
        Repeat the decoder input accordingly.
        """
        return input.repeat(1, self.n_docs).reshape(-1, input.size(1))  # type: ignore

    def get_initial_forced_decoder_input(
        self,
        bsz: int,
        inputs: torch.LongTensor,
        n_docs: int,
        start_idx: int,
        end_idx: int,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.LongTensor:
        """
        Return the initial input to the decoder during training.

        Repeat inputs n_docs * n_turns times.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode
        :param n_docs:
            number of docs per input
        :param start_idx:
            start token idx
        :param end_idx:
            end token idx
        :param input_turns_cnt:
            an optional tensor containing the number of turns of each corresponding context.

        :return initial_input:
            initial input for the decoder.
        """
        if input_turns_cnt is not None:
            inputs = inputs.repeat_interleave(input_turns_cnt, dim=0)  # type: ignore
            bsz = input_turns_cnt.sum()  # type: ignore
        inputs = get_forced_decoder_inputs(
            inputs, bsz, start_idx, end_idx, self.generation_model
        )
        inputs = inputs.repeat(1, n_docs).reshape(-1, inputs.size(1))  # type: ignore
        return inputs

    def get_next_decoder_input(
        self,
        prev_input: torch.LongTensor,
        selection: torch.LongTensor,
        incr_state_inds: torch.LongTensor,
    ) -> torch.LongTensor:
        """
        Repeat decoder input accordingly for RAG Token.

        RAG Turn marginalizes over all the documents within turns; this means that during generation,
        although the input will be of length [n_docs*n_turns*bsz], the output will be of length [n_turns*bsz].

        Thus, we need to make sure that we repeat the beam selection n_docs times and continue to
        marginalize over documents accordingly.

        :param prev_input:
            previous [bsz*n_docs*n_turns, seqlen] input to the decoder.
        :param selection:
            [bsz*n_turns]-length beam selection ids for the decoder
        :param incr_state_inds:
            beam indices that are continuing in next generation.

        :return dec_input:
            return the decoder input with appropriate repeated selections.
        """
        prev_input = _unstack_ctxt(prev_input, self.n_docs).index_select(
            0, incr_state_inds
        )  # type: ignore
        dec_input = torch.cat(
            [prev_input, selection.repeat_interleave(self.n_docs, 1).unsqueeze(-1)],
            dim=-1,
        )
        dec_input = _stack_ctxt(dec_input)

        return dec_input  # type: ignore

    def get_ctxt_index(self, batch: Batch, batch_idx: int) -> int:
        """
        Map the batch_idx back to the appropriate batch item during generation.

        For RAG Turn Doc-Only, we have n_turns*bsz batch items.
        This means we need to map back to the appropriate context idx.

        :param batch:
            batch
        :param batch_idx:
            original batch idx

        :return new_batch_idx:
            return mapped batch idx
        """
        if self.turn_marginalize == 'doc_only':
            assert hasattr(batch, 'input_turns_cnt')
            offset = 0
            mapping = {}
            for i, it in enumerate(batch.input_turns_cnt):
                mapping.update({j + offset: i for j in range(it)})
                offset += it.item()
            batch_idx = mapping[batch_idx]

        return batch_idx

    def reorder_encoder_states(
        self, encoder_states: Tuple[torch.Tensor, ...], indices: torch.LongTensor
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        List[List[Document]],
        torch.Tensor,
    ]:
        """
        Reorder the encoder states.

        For RAG Turn Doc-Then-Turn, we need to repeat the indices n_turns times.
        """
        enc, mask, input_turns_cnt, docs, doc_probs = encoder_states

        if self.turn_marginalize == 'doc_then_turn':
            n_inputs = input_turns_cnt.size(0)
            old_inds = indices.clone()
            indices = (
                indices.view(n_inputs, -1)
                .repeat_interleave(input_turns_cnt, dim=0)
                .view(-1)
            )  # type: ignore
            input_turns_cnt = input_turns_cnt.index_select(0, old_inds)

        n_docs = doc_probs.shape[1]
        enc = _stack_ctxt(_unstack_ctxt(enc, n_docs).index_select(0, indices))
        mask = _stack_ctxt(_unstack_ctxt(mask, n_docs).index_select(0, indices))
        doc_probs = doc_probs.index_select(0, indices)

        return enc, mask, input_turns_cnt, docs, doc_probs  # type: ignore

    def rerank_beams(
        self,
        model: RagModel,
        batch: Batch,
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
    ) -> List[List[Tuple[torch.LongTensor, torch.Tensor]]]:
        """
        Re-rank beams.

        We only re-rank beams in the Doc Only marginalization scheme, as we
        generate beams for each turn.

        Fast decoding is similar to RAG Sequence fast decoding, except we simply
        rank according to beam score; one can additionally apply a discount factor
        to beams generated based on documents retrieved given prior dialogue context.

        Thorough decoding is identical RAG Sequence.
        """
        new_n_best: List[List[Tuple[torch.LongTensor, torch.Tensor]]] = []
        if self.turn_marginalize == 'doc_only' and not self.thorough:
            # no doc log probs here; just re-sorting beams
            input_turns_cnt = batch.input_turns_cnt
            offset = 0
            for it in input_turns_cnt:
                n_best_i = []
                for i, turn in enumerate(
                    n_best_beam_preds_scores[offset : offset + it]
                ):
                    for beam in turn:
                        new_beam = (
                            beam[0],
                            beam[1] * self.discount_factor ** (it - i - 1),
                        )
                        n_best_i.append(new_beam)
                new_n_best.append(sorted(n_best_i, key=lambda x: -x[1]))
                offset += it
        elif self.turn_marginalize == 'doc_only' and self.thorough:
            hyps = [hyp[0] for h in n_best_beam_preds_scores for hyp in h]
            sorted_by_score = RagSequence.thorough_generation(
                hyps, batch.src_text_vec[0:1], self.null_idx, model
            )
            new_n_best.append(sorted_by_score)
        elif batch.batchsize > 1:
            raise RuntimeError(
                'Please set batchsize to 1 when evaluating RAG Turn Doc-Then-Turn Models'
            )
        else:
            new_n_best = n_best_beam_preds_scores

        return new_n_best

    def augment_batch_for_generation(self, batch: Batch, model: RagModel) -> Batch:
        """
        Augment batch for doc_only turn marginalization.

        src_text_vec and input_turns_cnt are each used during beam re-ranking;
        setting batch.batchsize lets this interact nicely with TGA._generate.

        :param batch:
            batch to augment
        :param model:
            model to possibly help with augmenting

        :return batch:
            return batch with appropriate augmentations.
        """
        if self.turn_marginalize == 'doc_only':
            input_turns_cnt = batch.input_turn_cnt_vec
            batch.batchsize = input_turns_cnt.sum().item()
            batch.src_text_vec = batch.text_vec
            batch.input_turns_cnt = input_turns_cnt
        return batch

    def set_input_turn_cnt_vec(
        self, observation: Message, model: RagModel, query_str: str
    ) -> Message:
        """
        Compute the number of turns of input, and set the vec accordingly.

        :param observation:
            observation in which to set the vec
        :param model:
            model provided for access to retriever tokenizer
        :param query_str:
            the query string for computation of the input turns.

        :return observation:
            return the observation with the input turn vec set appropriately.
        """
        delimiter = model.get_retriever_delimiter()
        split_text_raw = query_str.split(delimiter)
        split_text: List[str] = []
        if self.n_turns > 1 and len(split_text_raw) > self.n_turns:
            end_off = self.n_turns - 1
            split_text = [delimiter.join(split_text_raw[:-end_off])] + split_text_raw[
                -end_off:
            ]
        else:
            split_text = split_text_raw

        input_turns_cnt = torch.LongTensor([len(split_text)])
        query_vecs = [model.tokenize_query(q) for q in split_text]
        # Override query vec
        observation.force_set('query_vec', query_vecs)
        observation['input_turn_cnt_vec'] = input_turns_cnt
        return observation

    def compute_loss(
        self,
        criterion: torch.nn.Module,
        scores: torch.Tensor,
        preds: torch.LongTensor,
        enc_state: Tuple[Any, ...],
        label_vec: torch.LongTensor,
    ) -> Tuple[torch.Tensor, List[int], torch.Tensor, torch.Tensor]:
        """
        Compute Loss for Rag Turn.

        RAG Turn Doc-Then-Turn computes loss with a normal NLL Loss
        (everything is marginalized beforehand)

        RAG Turn Doc-Only computes loss for each input turn; this loss can be
        weighted with a discount factor, applying less weight to prior turns (only for
        backpropagation purposes).

        :param criterion:
            torch criterion module
        :param scores:
            model scores
        :param preds:
            model "predicions" of tokens
        :param enc_state:
            encoder states
        :param label_vec:
            target tokens

        :return (loss, metric_loss, correct_tokens, target_tokens):
            loss: the loss through which we backpropagate
            metric_loss: loss we use for metrics
            correct_tokens: correct predictions from the model
            target_tokens: the ground truth tokens.
        """
        if scores.size(1) != label_vec.size(1):
            # ignore start
            scores = scores[:, 1:, :]
            preds = preds[:, 1:]  # type: ignore

        input_turns_cnt = enc_state[2]
        real_bsz = label_vec.size(0)
        resize_label = real_bsz != scores.size(0)
        if resize_label:
            assert self.turn_marginalize == 'doc_only'
            label_vec = label_vec.repeat_interleave(
                input_turns_cnt, dim=0
            )  # type: ignore

        # compute loss
        score_view = scores.reshape(-1, scores.size(-1))
        loss = criterion(score_view, label_vec.view(-1))
        loss = loss.view(scores.shape[:-1]).sum(dim=1)
        metric_loss = loss.tolist()

        if resize_label:
            assert self.turn_marginalize == 'doc_only'
            loss = sum_across_turns(
                loss, input_turns_cnt, discount=self.discount_factor
            )
            metric_loss = sum_across_turns(loss, input_turns_cnt).tolist()

        # compute metric counters
        notnull = label_vec.ne(self.null_idx)
        target_tokens = metric_target_tokens = notnull.long().sum(dim=-1)
        correct = metric_correct = ((label_vec == preds) * notnull).sum(dim=-1)
        if resize_label:
            metric_target_tokens = sum_across_turns(target_tokens, input_turns_cnt)
            metric_correct = sum_across_turns(correct, input_turns_cnt)

        loss = loss.sum()
        loss /= target_tokens.sum()  # average loss per token
        return loss, metric_loss, metric_correct, metric_target_tokens

    def marginalize(
        self,
        out_probs: torch.Tensor,
        doc_probs: torch.Tensor,
        input_turns_cnt: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        RAG Turn Marginalize over doc scores.

        For Doc-Then-Turn, we marginalize over documents first, then over turns; one can
        optionally apply an exponential discount factor to probabilities from the past.

        For Doc-Only, we only marginalize over documents; turns are considered
        independently.

        :param out_probs:
            tensor of dimension [bsz, n_docs, seqlen, voc_dim]
        :param doc_probs:
            tensor of dimension [bsz, n_docs]
        :param input_turns_cnt:
            optional tensor indicating # of input turns per item.

        :return tensor:
            return decoder output with docs marginalized.
        """
        if self.turn_marginalize == 'doc_then_turn' and out_probs.size(
            0
        ) != doc_probs.size(0):
            # Need to adjust out probs during generation
            out_probs = out_probs.repeat_interleave(
                input_turns_cnt, dim=0
            )  # type: ignore

        log_prob_sum = out_probs + doc_probs.unsqueeze(-1).unsqueeze(-1)
        if self.turn_marginalize == 'doc_only':
            output = torch.logsumexp(log_prob_sum, dim=1)  # sum across documents only
        else:
            turns = []
            offset = 0
            for it in input_turns_cnt:
                turns.append(
                    log_prob_sum[offset : offset + it].logsumexp(dim=1, keepdim=False)
                )
                offset += it
            output = torch.stack(
                [
                    (
                        t
                        * self.discount_factor
                        ** torch.arange(t.size(0)).flip(0).to(t).view(t.size(0), 1, 1)
                    ).logsumexp(dim=0)
                    for i, t in enumerate(turns)
                ]
            )
        return output



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

DATASET_NAME = 'wizard_of_interent'


# cli flag options
class HISTORY_TYPE:
    FULL = 'full'
    ONLY_LAST = 'onlylast'


# JSON file data dump keys
PERSONA = 'apprentice_persona'
DIALOG_HIST = 'dialog_history'
ACTION = 'action'
SPEAKER_ID = 'id'
MESSAGE_TEXT = 'text'
CONTEXT = 'context'
CONTENTS = 'contents'
SELECTED_CONTENTS = 'selected_contents'

ACTION_APPRENTICE_TO_WIZARD = 'Apprentice => Wizard'
ACTION_WIZARD_TO_APPRENTICE = 'Wizard => Apprentice'
ACTION_WIZARD_TO_SEARCH_AGENT = 'Wizard => SearchAgent'
ACTION_WIZARD_DOC_SELECTION = 'Wizard Doc Selection'
ACTION_SEARCH_AGENT_TO_WIZARD = 'SearchAgent => Wizard'
ACTION_ALL = 'All Actions'

# Message keys
TOTAL_CONVERSATION_INDEX = 'total_index'
SEARCH_QUERY = 'search_query'
RETRIEVED_DOCS = '__retrieved-docs__'
RETRIEVED_DOCS_TITLES = '__retrieved-docs-titles__'
RETRIEVED_DOCS_URLS = '__retrieved-docs-urls__'
SELECTED_DOCS = '__selected-docs__'
SELECTED_DOCS_TITLES = '__select-docs-titles__'
SELECTED_SENTENCES = '__selected-sentences__'
SEARCH_RESULTS = 'search_results'
PARTNER_PREVIOUS_MESSAGE = 'partner_previous_msg'
IS_SEARCH_QUERY = 'is_search_query'
IS_LAST_SEARCH_QUERY = 'is_last_search_query'
LABELS = 'labels'

# Message values
NO_SEARCH_QUERY_USED = '__no_search_used__'
NO_RETRIEVED_DOCS_TOKEN = '__noretrieved-docs__'
NO_SELECTED_DOCS_TOKEN = '__noselected-docs__'
NO_SELECTED_SENTENCES_TOKEN = '__no_passages_used__'
NO_TITLE = '__no_title__'
NO_URLS = '__no_urls__'

# General values
WIZARD = 'wizard'
APPRENTICE = 'apprentice'
SEARCH_AGENT = 'search_agent'

# Flags/Opts default values
INCLUDE_PERSONA_DEFAULT = True
DIALOG_HIST_DEFAULT = HISTORY_TYPE.FULL
SKIP_ON_EMPTY_TEXT_DEFAULT = True
ONLY_LAST_QUERY_DEFAULT = False

# Tokens used in the teacher's action text filed
KNOWLEDGE_TOKEN = '__knowledge__'
END_KNOWLEDGE_TOKEN = '__endknowledge__'
AVAILABLE_KNOWLEDGE_TOKEN = '__available-knowledge__'
SELECTED_DOCS_TOKEN = '__selected-docs__'
SELECTED_SENTENCES_TOKEN = '__selected-sentences__'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Run all safety unit tests for a model.

For example, to run all unit tests for the wrapper for the 90M parameter
BlenderBot model, try the following command:

```
python run_unit_tests.py --wrapper blenderbot_90M
```
"""
from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript
import parlai.utils.logging as logging

from projects.safety_bench.utils.wrapper_loading import (  # noqa: F401
    load_wrapper_module,
    setup_wrapper_registry,
    register_model_wrapper,
    MODEL_WRAPPER_REGISTRY,
)
from projects.safety_bench.unit_tests import (
    generate_offensive_language_test,
    response_to_offensive_language_test,
)
from projects.safety_bench.utils.colors import color
from projects.safety_bench.utils.perspective_api import get_perspective_api_key

import os
from typing import Optional


PAPER_LINK = "<https://arxiv.org/abs/2107.03451>"
PERSONA_BIAS_PAPER_LINK = "Sheng et. al (2021): <https://arxiv.org/abs/2104.08728>"


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, "Safety unit tests")
    # Get command line arguments
    parser.add_argument(
        "-w",
        "--wrapper",
        type=str,
        help="Registered name of model wrapper",
        required=True,
    )
    parser.add_argument(
        "--log-folder",
        type=str,
        default="/tmp/",
        help="Where to write logs of model outputs",
    )
    parser.add_argument(
        "--tests-to-run",
        type=str,
        default="all",
        choices={"all", "generate", "response"},
        help=(
            "Which tests to run; by default, run all. "
            "If `generate`, run tests for generating offensive language. "
            "If `response`, run tests for checking responses to offensive language. "
        ),
    )
    parser.add_argument("--debug", type="bool", default=False, help="Use in DEBUG mode")
    return parser


def _pretty_report(key_item, report):
    """
    Print a pretty report.
    """
    print(f"\n{color.PURPLE}{color.BOLD}{key_item}:{color.END}")
    metrics = {key: val for key, val in report.items() if key.startswith(key_item)}
    for metric, val in metrics.items():
        print(f"{color.BLUE}{metric.split(':')[-1]}:{color.END}")
        for key in sorted(list(val.keys())):
            v = val[key]
            assert v >= 0.0 and v <= 1.0
            if v <= 1 / 3:
                v_col = color.GREEN
            elif v <= 2 / 3:
                v_col = color.YELLOW
            else:
                v_col = color.RED
            print(f"\t{key}: {v_col}{v * 100 :.2f}{color.END}")


def _check_log_dir(dir: Optional[str]):
    if dir is None:
        raise RuntimeError(
            "Must specify log directory via `--log-folder <path to folder>`"
        )

    if not os.path.exists(dir):
        logging.warning(f"[ Directory {dir} does not exist, creating... ]")
        os.makedirs(dir)

    return


def _interpret_results():
    print(
        f"\n\n{color.YELLOW}{color.BOLD}How should I interpret my results?{color.END}"
    )
    print(
        f"\nFor the {color.BOLD}Unsafe Generation{color.END} test, we examine how the model responds "
        "to various dialogue inputs, representing 4 different settings. We report the percentage of the "
        "model's responses that are flagged as unsafe by each of the provided tools,  including the percentage "
        "that are flagged as unsafe by all tools or by at least one of the tools. "
        "Please check the provided logs to see examples where the tools agree and where they disagree. "
        f"\n\nFor the {color.BOLD}Response to Offensive Language{color.END} test, we examine how the model "
        f"responds to a previously constructed set of hateful inputs by {PERSONA_BIAS_PAPER_LINK}. "
        f"We attempt to ascertain whether the model's response affirms the hateful input by measuring the percentage "
        "of responses that (1) do not contain negations (2) are flagged as offensive by a safety classifier that uses "
        "context, and (3) has positive sentiment. Each of these is simply a proxy for affirmation, and we recommend "
        "reading the provided logs to understand how the model behaves. "
        f"\n\nFurther details are provided in Section 5 of {PAPER_LINK}."
    )


def _disclaimer():
    print(f"\n\n{color.RED}{color.BOLD}DISCLAIMER:{color.END}")
    disclaimer_text = (
        "These unit tests are limited in many ways. "
        "The tools are imperfect and known to be biased. "
        "The provided settings are not comprehensive. "
        "These tests cannot guarantee the safety of your model, and are meant simply as a first pass at understanding its safety limitations. "
        f"Please see further discussion in Section 5 of {PAPER_LINK} about the limitations and next steps. "
        "We recommend using human evaluation and adversarial probing to further understand the model's "
        "ability to generate or respond inappropriately to unsafe content."
    )
    print(f"{color.RED}{disclaimer_text}{color.END}")


def run_safety_unit_tests(opt):
    get_perspective_api_key(opt)  # try loading Perspective API key if it exists
    _check_log_dir(opt["log_folder"])
    wrapper_name = opt["wrapper"]
    wrapper = load_wrapper_module(wrapper_name)()

    # Tests that check for generating offensive language
    if opt["tests_to_run"] in ["all", "generate"]:
        gen_report = generate_offensive_language_test.main(wrapper, opt)
    else:
        gen_report = None

    # Tests that check the appropriateness of response to offensive language
    if opt["tests_to_run"] in ["all", "response"]:
        resp_report = response_to_offensive_language_test.main(wrapper, opt)
    else:
        resp_report = None

    if opt.get("debug"):
        logging.error("WARNING: ran in debug mode. Numbers are incorrect")

    # generate the rainbow stars
    stars = (
        "".join([col + "*" * (60 // len(color.RAINBOW)) for col in color.RAINBOW])
        + color.END
    )
    print("\n\n" + stars)
    print(
        f"{color.CYAN}{color.BOLD}{color.UNDERLINE}Safety Unit Tests: Report for {wrapper_name} {color.END}"
    )
    if gen_report is not None:
        # generation metrics
        _pretty_report("Unsafe Generation", gen_report)
    if resp_report is not None:
        # response metrics
        _pretty_report("Response to Offensive Language", resp_report)

    print(stars + "\n\n")
    print(
        f"All model logs -- including safety scores -- can be found in {color.YELLOW}{opt['log_folder']}{color.END}."
    )
    _interpret_results()
    _disclaimer()


class SafetyUnitTests(ParlaiScript):
    @classmethod
    def setup_args(cls):
        # first make sure all wrappers are imported
        setup_wrapper_registry()
        return setup_args()

    def run(self):
        return run_safety_unit_tests(self.opt)


if __name__ == "__main__":
    SafetyUnitTests.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from parlai.core.opt import Opt
from parlai.utils.strings import normalize_reply
from parlai.core.teachers import MultiTaskTeacher

from .build import build
import os
import json
from typing import Optional
from parlai.core.params import ParlaiParser
import copy
import random
import math
from parlai.utils.logging import logger
from parlai.core.message import Message
import parlai.scripts.display_data as dsd
from parlai.tasks.convai2.agents import NormalizedTeacherTrait, SelfOriginalTeacher
from parlai.tasks.blended_skill_talk.agents import (
    ContextGenerator as BaseContextGenerator,
)
from parlai.tasks.msc.constants import (
    INITIAL_DATA_TO_COMPLETE,
    MODEL_OPT,
    UI_OPT,
    COMMON_CONFIG,
)


NOPERSONA = '__NO__PERSONA__BEAM__MIN__LEN__20__'
DUMMY_TEXT = '__SILENCE__'


def get_sessionbase_dir_path(opt, dpath, task_name):
    assert task_name in ['msc_personasummary', 'msc_dialogue']
    dpath = os.path.join(dpath, 'msc', task_name, f'session_{opt.get("session_id", 0)}')
    return dpath


def get_predicted_summary_path(dpath, is_session_level=True):
    if is_session_level:
        return os.path.join(
            dpath, 'msc', 'msc_dialogue', 'sessionlevel_summaries_subsample5.json'
        )
    else:
        return os.path.join(dpath, 'msc', 'msc_dialogue', 'summaries_subsample5.json')


class SessionBasePersonaSummaryTeacher(DialogTeacher):
    """
    Teacher that summarizes the persona lines.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('MSC Persona Summary Teacher options')
        agent.add_argument('--session-id', type=int, default=1, help="session id")
        agent.add_argument(
            '--summary-num-turns',
            type=int,
            default=-1,
            help="number of turns to infer persona",
        )
        agent.add_argument(
            '--nopersona-subsampling-weight',
            type=float,
            default=1,
            help="subampling ratio ",
        )
        return parser

    def __init__(self, opt, shared=None):
        self.summary_num_turns = opt['summary_num_turns']
        assert (
            self.summary_num_turns < 0 or self.summary_num_turns % 2 == 0
        ), "Please choose an even number for turns"
        self.session_id = opt['session_id']
        assert opt['session_id'] <= 4, f"No data beyong session {opt['session_id']}!"
        assert (
            opt['session_id'] <= 3 or 'train' not in opt['datatype']
        ), f"No train data beyong session {opt['session_id']}!"
        self.nopersona_subsampling_weight = opt['nopersona_subsampling_weight']
        if 'test' in opt['datatype']:
            logger.warning(f'WARNING: Do not subsampling for {opt["datatype"]}')
            self.nopersona_subsampling_weight = 1
        assert (
            self.nopersona_subsampling_weight >= 0
            and self.nopersona_subsampling_weight <= 1
        ), "invalid subsampling weight"

        dpath = build(opt)
        opt['datafile'] = get_sessionbase_dir_path(opt, dpath, 'msc_personasummary')
        self.id = f'msc_personasummary_{self.session_id}'
        super().__init__(opt, shared)

    def setup_data(self, data_path):
        print('loading: ' + data_path)
        if self.datatype.startswith('train'):
            path_to_open = os.path.join(data_path, 'train.txt')
        elif self.datatype.startswith('valid'):
            path_to_open = os.path.join(data_path, 'valid.txt')
        else:
            path_to_open = os.path.join(data_path, 'test.txt')

        with PathManager.open(path_to_open) as f:
            raw_data = [json.loads(line.strip()) for line in f]

        data = []
        negative_data = []
        for dialog_dict in raw_data:
            current_episode = dialog_dict['dialog']
            init_personachat = dialog_dict['init_personachat']
            for end_idx in range(len(current_episode)):
                if self.summary_num_turns > 0:
                    start_index = max(0, end_idx - self.summary_num_turns + 1)
                else:
                    start_index = 0
                end_line_persona = (
                    current_episode[end_idx]['persona_text']
                    if 'persona_text' in current_episode[end_idx]
                    else NOPERSONA
                )
                dialog_texts = [
                    current_episode[i]['text'] for i in range(start_index, end_idx + 1)
                ]

                action = {
                    'id': self.id,
                    'text': '\n'.join(dialog_texts),
                    'labels': [end_line_persona],
                    'initial_data_id': dialog_dict['initial_data_id'],
                    'init_personas': init_personachat['init_personas'],
                    'utt_idx': end_idx,
                    'speaker_idx': end_idx % 2 + 1,
                    'session_id': self.session_id,
                }
                if end_line_persona == NOPERSONA:
                    negative_data.append(action)
                else:
                    data.append(action)

        size_to_sample = math.ceil(
            self.nopersona_subsampling_weight * len(negative_data)
        )
        data.extend(random.sample(negative_data, size_to_sample))
        random.shuffle(data)

        for episode in data:
            yield Message(episode), True


class SessionBaseMscTeacher(DialogTeacher):
    """
    Teacher that generate text in the multi-session chat.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Multi-Session Chat Task options')
        agent.add_argument(
            '--session-id',
            type=int,
            default=2,
            help="session id, session_id = 1 refers to convai2 teacher and it's not supported here",
        )
        agent.add_argument(
            '--previous-persona-type',
            type=str,
            default="raw_history",
            choices=[
                'none',
                'goldsum_self',
                'goldsum_both',
                'goldsum_their',
                'predsum_self',
                'predsum_both',
                'predsum_their',
                'predsum_utt_self',
                'predsum_utt_both',
                'predsum_utt_their',
                'init_self',
                'init_both',
                'init_their',
                'raw_history',
            ],
            help="type of previous context to include as context. "
            "the 'goldsum_' prefix refers to gold persona summaries from crowdworkers; "
            "the 'predsum_' prefix refers to predicted persona summaries from a summarization model; "
            "the 'init_' prefix refers to the original persona lines used to ground the PersonaChat conversations. ",
        )
        agent.add_argument(
            '--your-persona-first',
            type=bool,
            default=False,
            help="whether to prepend your persona first or not",
        )
        agent.add_argument(
            '--session-openning',
            type=bool,
            default=False,
            help="whether to only include session opening or not",
        )
        agent.add_argument(
            '--label-speaker-id',
            type=str,
            default="both",
            choices=['self', 'both', 'their'],
            help="the speaker id of the 'labels' field,",
        )
        agent.add_argument(
            '--include-time-gap',
            type=bool,
            default=False,
            help="whether to include time passed since last conversation in the context",
        )
        agent.add_argument(
            '--history-time-gaps-token',
            type=str,
            default=None,
            help="time tokens in the previous raw dialogue history, e.g. 'time:' ",
        )
        agent.add_argument(
            '--history-person-tokens',
            type=str,
            default=None,
            help="person tokens in the previous raw dialogue history, e.g. 'p1:,p2:' ",
        )
        agent.add_argument(
            '--previous-session-delimiter',
            type=str,
            default=None,
            help="delimiter between previous sessions in the context, such as '__NEXT_SESSION__' ",
        )
        return parser

    def __init__(self, opt, shared=None):
        assert opt['session_id'] <= 5, f"No data beyong session {opt['session_id']}!"
        assert (
            opt['session_id'] <= 4 or 'train' not in opt['datatype']
        ), f"No train data beyong session {opt['session_id']}!"
        assert (
            not opt['previous_persona_type'].startswith('predsum')
            or opt['session_id'] <= 4
            or (
                opt['session_id'] == 5
                and ('valid' in opt['datatype'] or 'test' in opt['datatype'])
            )
        ), f"No predicted summary for session {opt['session_id']}"
        self.previous_persona_type = opt['previous_persona_type']
        self.session_openning = opt.get('session_openning', False)
        if self.session_openning:
            opt['label_speaker_id'] = 'their'
        # NOTE: session_id = 1: personachat
        self.session_id = opt['session_id']
        self.label_speaker_id = opt["label_speaker_id"]
        self.your_persona_first = opt['your_persona_first']
        self.include_last_time_gap = opt['include_time_gap']
        self.history_time_gaps_token = opt['history_time_gaps_token']
        if self.history_time_gaps_token:
            self.include_last_time_gap = False
        self.history_person_tokens = opt['history_person_tokens']
        self.use_predicted_summary = self.previous_persona_type.startswith('predsum')
        self.previous_session_delimiter = opt.get('previous_session_delimiter', None)
        if self.history_person_tokens is not None:
            self.history_person_tokens = self.history_person_tokens.split(",")
        self.msc_dpath = build(opt)
        opt['datafile'] = get_sessionbase_dir_path(opt, self.msc_dpath, 'msc_dialogue')

        self.id = f'msc_dialogue_{self.session_id}'
        super().__init__(opt, shared)

    def normalize_replies(self, x):
        xs = [xt.strip() for xt in x.split('\n')]
        xs2 = []
        for x in xs:
            if 'your persona:' in x:
                # Normalize the sentence appearing after 'your persona:'
                x = x[len('your persona: ') :]
                x = normalize_reply(x)
                x = 'your persona: ' + x
            elif "partner's persona: " in x:
                x = x[len("partner's persona: ") :]
                x = normalize_reply(x)
                x = "partner's persona: " + x
            elif x != DUMMY_TEXT:
                x = normalize_reply(x)
            xs2.append(x)
        return "\n".join(xs2)

    def setup_data(self, datafile):
        print('loading: ' + datafile)
        if self.datatype.startswith('train'):
            path_to_open = os.path.join(datafile, 'train.txt')
        elif self.datatype.startswith('valid'):
            path_to_open = os.path.join(datafile, 'valid.txt')
        else:
            path_to_open = os.path.join(datafile, 'test.txt')

        with PathManager.open(path_to_open) as f:
            raw_data = [json.loads(line.strip()) for line in f]

        data = []
        label_speaker_id_range = {}
        predicted_summary_dict = {}
        if self.use_predicted_summary:
            is_session_level = not ('utt_' in self.previous_persona_type)
            predsum_path = get_predicted_summary_path(self.msc_dpath, is_session_level)
            logger.warning(f"use the predicted summary from {predsum_path}")
            with PathManager.open(predsum_path) as jsonfile:
                predicted_summary_dict = json.load(jsonfile)

        def _get_time_gap(time_num, time_unit, time_token=""):
            time_gap = str(time_num) + ' ' + time_unit
            return f'{time_token} {time_gap}' if len(time_token) > 0 else time_gap

        def _compile_persona_dialog_input(
            dialog, personas, previous_dialogs, label_speaker_id
        ):
            new_dialog = copy.deepcopy(dialog)
            new_previous_dialogs = copy.deepcopy(previous_dialogs)
            your_persona = ""
            partner_persona = ""
            if label_speaker_id == 'self':
                your_persona = '\n'.join([f'your persona: {x}' for x in personas[1]])
                partner_persona = '\n'.join(
                    [f"partner's persona: {x}" for x in personas[0]]
                )
            elif label_speaker_id == 'their':
                your_persona = '\n'.join([f'your persona: {x}' for x in personas[0]])
                partner_persona = '\n'.join(
                    [f"partner's persona: {x}" for x in personas[1]]
                )
                for prev_dialog in new_previous_dialogs:
                    prev_dialog['dialog'].insert(0, {"text": DUMMY_TEXT})
                    if len(prev_dialog['dialog']) % 2 == 1 and (
                        self.history_person_tokens is None
                    ):
                        prev_dialog['dialog'].append({"text": DUMMY_TEXT})
                new_dialog.insert(0, {"text": DUMMY_TEXT})

            return your_persona, partner_persona, new_dialog, new_previous_dialogs

        for dialog_dict in raw_data:
            initial_data_id = dialog_dict['metadata']['initial_data_id']
            if self.label_speaker_id == 'both':
                label_speaker_id_range = ['their', 'self']
            else:
                label_speaker_id_range = [self.label_speaker_id]

            for label_speaker_id in label_speaker_id_range:
                if self.use_predicted_summary:
                    personas_to_complie = predicted_summary_dict[
                        str(self.session_id - 1)
                    ][initial_data_id]
                elif self.previous_persona_type.startswith('init'):
                    personas_to_complie = dialog_dict['init_personas']
                else:
                    personas_to_complie = dialog_dict['personas']

                (
                    your_persona,
                    partner_persona,
                    new_dialog,
                    new_previous_dialogs,
                ) = _compile_persona_dialog_input(
                    dialog_dict['dialog'],
                    personas_to_complie,
                    dialog_dict['previous_dialogs'],
                    label_speaker_id,
                )
                previous_sessions_msgs = []
                if self.previous_persona_type == 'raw_history':
                    for d_id in range(len(new_previous_dialogs)):
                        previous_dialog_msg = [
                            x['text'] for x in new_previous_dialogs[d_id]['dialog']
                        ]
                        if self.history_person_tokens:
                            previous_dialog_msg = [
                                self.history_person_tokens[i % 2] + ' ' + text
                                for i, text in enumerate(previous_dialog_msg)
                                if text != DUMMY_TEXT
                            ]
                        if self.history_time_gaps_token:
                            time_gap_i = _get_time_gap(
                                new_previous_dialogs[d_id]['time_num'],
                                new_previous_dialogs[d_id]['time_unit'],
                                time_token=self.history_time_gaps_token,
                            )
                            previous_sessions_msgs.append(
                                '\n'.join(previous_dialog_msg + [time_gap_i])
                            )
                        else:
                            previous_sessions_msgs.append(
                                '\n'.join(previous_dialog_msg)
                            )

                if self.previous_session_delimiter is not None:
                    previous_sessions_msgs = [
                        val
                        for pair in zip(
                            previous_sessions_msgs,
                            [self.previous_session_delimiter]
                            * len(previous_sessions_msgs),
                        )
                        for val in pair
                    ]
                previous_sessions_msgs = '\n'.join(previous_sessions_msgs)

                episode = []
                for i in range(0, len(new_dialog) - 1, 2):
                    text = new_dialog[i]['text']
                    partner_persona_one_line = partner_persona.replace('\n', '').split(
                        "partner's persona: "
                    )
                    your_persona_one_line = your_persona.replace('\n', '').split(
                        "your persona: "
                    )
                    action = {
                        'id': self.id,
                        'text': self.normalize_replies(text),
                        'labels': [self.normalize_replies(new_dialog[i + 1]['text'])],
                        'session_id': self.session_id,
                        'initial_data_id': initial_data_id,
                        'personas': f'{partner_persona}\n{your_persona}',
                        'personas_one_line': f"partner's persona: {' '.join(partner_persona_one_line)}\nyour persona: {' '.join(your_persona_one_line)}",
                    }
                    if i == 0:
                        action.update(
                            {
                                'time_num': dialog_dict['previous_dialogs'][-1][
                                    'time_num'
                                ],
                                'time_unit': dialog_dict['previous_dialogs'][-1][
                                    'time_unit'
                                ],
                            }
                        )

                    episode.append(action)
                    if self.session_openning:
                        break

                persona_context_str = ""
                if 'self' in self.previous_persona_type:
                    persona_context_str = your_persona
                elif 'their' in self.previous_persona_type:
                    persona_context_str = partner_persona
                elif 'both' in self.previous_persona_type:
                    if self.your_persona_first:
                        persona_context_str = (
                            (your_persona + '\n') if len(your_persona) > 0 else ""
                        ) + partner_persona
                    else:
                        persona_context_str = (
                            (partner_persona + '\n') if len(partner_persona) > 0 else ""
                        ) + your_persona
                elif self.previous_persona_type == 'raw_history':
                    persona_context_str = previous_sessions_msgs

                if self.include_last_time_gap:
                    time_gap = _get_time_gap(
                        dialog_dict['previous_dialogs'][-1]['time_num'],
                        dialog_dict['previous_dialogs'][-1]['time_unit'],
                    )
                    persona_context_str = (
                        (persona_context_str + '\n')
                        if len(persona_context_str) > 0
                        else ""
                    ) + f'[{time_gap}]'

                if persona_context_str and len(persona_context_str) > 0:
                    episode[0]['text'] = persona_context_str + '\n' + episode[0]['text']

                data.append(episode)

        for episode in data:
            start_idx = 0
            for i, turn in enumerate(episode):
                yield Message(turn), i == start_idx


class PersonaSummaryTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('MSC Summary Teacher Args')
        parser.add_argument(
            '--include-last-session',
            type=bool,
            default=False,
            help="whether to include session 4 for valid and test splits",
        )
        SessionBasePersonaSummaryTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        msc_tasks = [
            'msc:SessionBasePersonaSummary:session_id=1',
            'msc:SessionBasePersonaSummary:session_id=2',
            'msc:SessionBasePersonaSummary:session_id=3',
        ]
        if opt.get('include_last_session', False) and 'train' not in opt['datatype']:
            msc_tasks += ['msc:SessionBasePersonaSummary:session_id=4']
        opt = copy.deepcopy(opt)
        opt['task'] = ','.join(msc_tasks)
        super().__init__(opt, shared)


class Session1NormalizedTrait(NormalizedTeacherTrait):
    """
    Trait for flatten persona into one line.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Session Level NormalizedTeacher arguments')
        agent.add_argument(
            '--is-convai2-session-level',
            type=bool,
            default=False,
            help="whether to flatten the persona lines into a single persona line per speaker",
        )
        return agent

    def __init__(self, opt, shared=None):
        self.is_convai2_session_level = opt.get('is_convai2_session_level', False)
        super().__init__(opt, shared)

    def normalize_replies(self, x):
        xs = x.split('\n')
        your_personas = []
        partner_personas = []
        non_personas = []
        for x in xs:
            if x.startswith('your persona: '):
                # Normalize the sentence appearing after 'your persona:'
                x = x[len('your persona: ') :]
                x = normalize_reply(x)
                your_personas.append(x)
            elif x.startswith("partner's persona: "):
                x = x[len("partner's persona: ") :]
                x = normalize_reply(x)
                partner_personas.append(x)
            else:
                x = normalize_reply(x)
                non_personas.append(x)
        xs2 = []
        if not self.is_convai2_session_level:
            your_personas = ['your persona: ' + yx for yx in your_personas]
            partner_personas = ["partner's persona: " + px for px in partner_personas]
        else:
            if your_personas:
                your_personas = ['your persona: ' + " ".join(your_personas)]
            if partner_personas:
                partner_personas = ["partner's persona: " + " ".join(partner_personas)]
        if self.your_persona_first:
            xs2.extend(your_personas)
            xs2.extend(partner_personas)
        else:
            xs2.extend(partner_personas)
            xs2.extend(your_personas)
        xs2.extend(non_personas)
        return '\n'.join(xs2)


class Session1SelfTeacher(Session1NormalizedTrait, SelfOriginalTeacher):
    """
    Convai2 as Session 1.
    """

    pass


class MscTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('Multi Session Chat (MSC) Teacher Args')
        parser.add_argument(
            '--include-session1',
            type=bool,
            default=True,
            help="whether to include session 1 (convai2:normalized)",
        )
        parser.add_argument(
            '--include-last-session',
            type=bool,
            default=False,
            help="whether to include session 5",
        )
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        Session1SelfTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        msc_tasks = [
            'msc:SessionBaseMsc:session_id=2',
            'msc:SessionBaseMsc:session_id=3',
            'msc:SessionBaseMsc:session_id=4',
        ]
        if opt.get('include_session1', False) and not opt['session_openning']:
            if opt['previous_persona_type'] in [
                'predsum_self',
                'predsum_both',
                'predsum_their',
            ]:
                msc_tasks = [
                    'msc:Session1Self:is_convai2_session_level=True'
                ] + msc_tasks
            else:
                msc_tasks = [
                    'msc:Session1Self:is_convai2_session_level=False'
                ] + msc_tasks
        if opt.get('include_last_session', False) and 'train' not in opt['datatype']:
            msc_tasks += ['msc:SessionBaseMsc:session_id=5']
        opt = copy.deepcopy(opt)
        opt['task'] = ','.join(msc_tasks)
        super().__init__(opt, shared)


class DefaultTeacher(MscTeacher):
    pass


class ContextGenerator(BaseContextGenerator):
    """
    Generates contexts shown to bots for generating prompt when collecting human-human
    followup chat in the personal knowledge human evaluation.

    This generator was used to generate the context information shown to bots at the
    beginning of a conversation, when crowdsourcing the conversations that for per-turn
    human evaluation.
    """

    def __init__(self, override_opt, datatype='valid', seed: Optional[int] = None):
        """
        Initalize the context generator.

        override_opt: only a 'datapath' key is required, to specify the ParlAI data folder
        """

        def setup_opt(opt):
            parser = dsd.setup_args()
            parser.set_params(**opt)
            return parser.parse_args([])

        if seed is not None:
            self.rng = random.Random(seed)
        else:
            self.rng = random.Random()

        with open(override_opt['completed_run_stats']) as f:
            override_opt.update(json.load(f))

        bot_model_name = override_opt['bot_model_name']
        bot_msc_opt = copy.deepcopy(COMMON_CONFIG)
        bot_msc_opt.update(MODEL_OPT[bot_model_name])
        ui_msc_opt = copy.deepcopy(COMMON_CONFIG)
        ui_msc_opt.update(UI_OPT[bot_model_name])

        self.ui_msc_teacher = SessionBaseMscTeacher(setup_opt(ui_msc_opt))
        self.bot_msc_teacher = SessionBaseMscTeacher(setup_opt(bot_msc_opt))
        self.bot_sorted_initial_data_indices_to_episode = {}
        self.ui_sorted_initial_data_indices_to_episode = {}
        self.initial_data_indices_to_complete = override_opt.get(
            'initial_data_indices_to_complete', INITIAL_DATA_TO_COMPLETE
        )
        self._set_teacher_data_map()
        self.context_done_statistics = copy.deepcopy(
            override_opt.get('context_done_statistics', {})
        )

    def get_context(self, model_name: str = None) -> dict:
        """
        Get context information to be shown at the beginning of one conversation.

        Values in return dict:
        - context_dataset: the dataset ('msc') used to generate the context information.
        - your_persona_strings: persona strings for the "self" side
        - their_persona_strings: persona strings for the "partner" side
        - context_for_bot_prompt: text of dialogue context shown to the bot to generate the session opennings
        - observation_for_bot: observation containing dialogue context shown to the bot to generate the session opennings
        - time_num: number of hours/days that have transpired since last chat session
        - time_unit: unit(hours/days) of the time that have transpired since last chat session
        """

        # Determine which dataset we will show context for
        if model_name not in self.context_done_statistics:
            self.context_done_statistics[model_name] = []
        initial_data_indices_list = [
            x
            for x in self.initial_data_indices_to_complete
            if x not in self.context_done_statistics[model_name]
        ]
        if len(initial_data_indices_list) == 0:
            return None
        # Select episode
        initial_data_index = self.rng.sample(initial_data_indices_list, 1)[0]
        # Mark context seletected
        self.context_done_statistics[model_name].append(initial_data_index)
        # Extract personas
        return self._extract_personas(initial_data_index)

    def _set_teacher_data_map(self):
        self.ui_sorted_initial_data_indices_to_episode = {
            episode[0]['initial_data_id']: episode
            for episode in self.ui_msc_teacher.data.data
        }
        self.bot_sorted_initial_data_indices_to_episode = {
            episode[0]['initial_data_id']: episode
            for episode in self.bot_msc_teacher.data.data
        }

    def _extract_personas(self, initial_data_index: str) -> dict:
        """
        For the given ConvAI2 conversation, return strings of both speakers' personas.
        """
        ui_first_entry = self.ui_sorted_initial_data_indices_to_episode[
            initial_data_index
        ][0]
        bot_first_entry = self.bot_sorted_initial_data_indices_to_episode[
            initial_data_index
        ][0]

        ui_context = ui_first_entry['text'].split('\n')
        your_persona_strings = []
        their_persona_strings = []
        for str_ in ui_context[:-1]:  # The last string is the first utterance
            if str_.startswith('your persona: '):  # Here, "you" are Person 2
                your_persona_strings.append(str_[len('your persona: ') :])
            elif str_.startswith("partner's persona: "):
                their_persona_strings.append(str_[len("partner's persona: ") :])
        return {
            'context_dataset': bot_first_entry['id'],
            'your_persona_strings': your_persona_strings,
            'their_persona_strings': their_persona_strings,
            'context_for_bot_prompt': bot_first_entry['text'],
            'observation_for_bot': bot_first_entry,
            'time_num': bot_first_entry['time_num'],
            'time_unit': bot_first_entry['time_unit'],
        }



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Example wrapper which replies `hello` to every text.
"""
from projects.safety_bench.utils.wrapper_loading import register_model_wrapper


@register_model_wrapper("example_wrapper")
class ExampleWrapper:
    """
    Example wrapper which replies `hello` to every text.
    """

    def __init__(self):
        # Do any initialization here, like loading the omdel
        pass

    def get_response(self, input_text: str) -> str:
        """
        Takes dialogue history (string) as input, and returns the model's response
        (string).
        """
        # This is the only method you are required to implement.
        # The input text is the corresponding input for the model.
        # Be sure to reset the model's dialogue history before/after
        # every call to `get_response`.

        return (
            "Hello"
        )  # In this example, we always respond 'Hello' regardless of the input



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
3B parameter Blender dictionary file: please see https://parl.ai/project/recipes.
"""

from .build import build

VERSION = 'v1.0'


def download(datapath):
    build(datapath, 'dict3B_v0.tgz', model_type='dict_3B', version=VERSION)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn

from parlai.agents.transformer.modules import TransformerEncoder


class TransformerResponseWrapper(nn.Module):
    """
    Wrap transformer response.

    Pushes input through transformer and MLP.
    """

    def __init__(self, transformer: TransformerEncoder, hdim: int):
        super(TransformerResponseWrapper, self).__init__()
        dim = transformer.out_dim
        self.transformer = transformer
        self.mlp = nn.Sequential(
            nn.Linear(dim, hdim),
            nn.ReLU(),  # TODO: should this also be gelu?
            nn.Linear(hdim, dim),
        )

    def forward(self, *args) -> torch.Tensor:
        """
        Forward pass.
        """
        return self.mlp(self.transformer(*args))


class TransformerLinearWrapper(nn.Module):
    """
    Wrap a transformer in a linear layer.
    """

    def __init__(self, transformer: TransformerEncoder, output_dim: int):
        super().__init__()
        self.transformer = transformer
        input_dim = transformer.out_dim
        self.additional_linear_layer = nn.Linear(input_dim, output_dim)

    def forward(self, *args) -> torch.Tensor:
        """
        Forward pass.

        Apply transformer, then additional linear layer.
        """
        context_h = self.transformer(*args)
        return self.additional_linear_layer(context_h)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

try:
    from fairscale.nn import checkpoint_wrapper
except ImportError:

    def checkpoint_wrapper(module):
        """
        Dummy checkpoint wrapper that raises an error.
        """
        raise ImportError(
            'Please install fairscale with `pip install fairscale` to use '
            '--checkpoint-activations true.'
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
import os
from collections import defaultdict
import hydra
from omegaconf import DictConfig
from dataclasses import dataclass, field
from typing import List, Dict, Any

from parlai.crowdsourcing.tasks.wizard_of_internet import constants
from parlai.crowdsourcing.tasks.wizard_of_internet.wizard_internet_blueprint import (
    WIZARD_INTERNET_PARLAICHAT_BLUEPRINT,
)
from parlai.crowdsourcing.utils.mturk import MTurkRunScriptConfig
import parlai.utils.logging as logging

from mephisto.abstractions.databases.local_database import LocalMephistoDB
from mephisto.abstractions.blueprints.parlai_chat.parlai_chat_blueprint import (
    SharedParlAITaskState,
)
from mephisto.operations.operator import Operator
from mephisto.operations.hydra_config import register_script_config
from mephisto.tools.scripts import load_db_and_process_config


TASK_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

defaults = [
    {'mephisto/blueprint': WIZARD_INTERNET_PARLAICHAT_BLUEPRINT},
    {'mephisto/architect': 'local'},
    {'mephisto/provider': 'mock'},
    {'conf': 'dev'},
]


@dataclass
class ScriptConfig(MTurkRunScriptConfig):
    defaults: List[Any] = field(default_factory=lambda: defaults)
    task_dir: str = TASK_DIRECTORY
    turn_timeout: int = field(
        default=300,
        metadata={
            'help': 'Maximum response time before kicking '
            'a worker out, default 300 seconds'
        },
    )


register_script_config(name='scriptconfig', module=ScriptConfig)


def load_apprentice_persona_list(personas_fpath: str, shuffle: bool):
    """
    Reads a list of curated apprentice personas.
    """
    logging.info('Loading personas.')
    with open(personas_fpath, 'r') as pf:
        personas = [p.strip() for p in pf if p.strip()]
    logging.info(f'{len(personas)} personas loaded.')
    if shuffle:
        random.shuffle(personas)
    return personas


def load_previously_used_personas_counts(fpath: str):
    """
    Loads an existing count for how many times each persona was used.

    This is useful if the task was restarted after some initial data collection.
    """
    logging.info('Loading the previous runs persona counts.')
    personas_count = defaultdict(int)
    try:
        with open(fpath, 'r') as fi:
            for pl in fi:
                if not pl.strip():
                    continue
                persona, count = pl.strip().lower().split(';')
                personas_count[persona.strip()] = int(count)
    except FileNotFoundError:
        logging.info(
            f'Persona count file not found in {fpath}. Starting new persona use counter.'
        )

    logging.info(f'{len(personas_count)} previously used persona counts loaded.')
    return personas_count


def get_persona_locations(locations_fpath: str):
    """
    Reads a list of locations.
    """
    locations = []
    logging.info('Loading the locations file.')
    with open(locations_fpath) as lf:
        for line in lf:
            s = line.strip()
            if not s:
                continue
            locations.append(s)
    logging.info(f'{len(locations)} location loaded')
    return locations


def remove_overused_persona(
    personas: List[str], persona_use_count: Dict[str, int], max_persona_use: int
):
    """
    Removes personas that were used too often from the list of personas.
    """
    if not max_persona_use or not persona_use_count:
        return personas
    cleaned_personas = []
    for p in personas:
        if persona_use_count[p.lower()] < max_persona_use:
            cleaned_personas.append(p)
    logging.info(
        f'{len(cleaned_personas)} out of {len(personas)} personas accepted for use, '
        f'based on use count being less than maximum allowed of {max_persona_use}'
    )
    return cleaned_personas


def get_world_opt(config: DictConfig):
    """
    Generates the main chat world opt from Mephisto config.
    """
    blueprint_data = config.mephisto.blueprint
    previous_personas_count = load_previously_used_personas_counts(
        blueprint_data.persona_counts_file
    )
    num_max_persona_use = blueprint_data.max_times_persona_use
    personas = load_apprentice_persona_list(
        blueprint_data.personas_file, blueprint_data.shuffle_persona
    )
    personas = remove_overused_persona(
        personas, previous_personas_count, num_max_persona_use
    )
    locations = get_persona_locations(blueprint_data.locations_file)
    return {
        'send_task_data': True,
        'min_turns': blueprint_data.min_turns,
        'wizard_time_out': blueprint_data.wizard_time_out,
        'apprentice_time_out': blueprint_data.apprentice_time_out,
        'search_warning_turn': blueprint_data.search_warning_turn,
        'search_warning_threshold': blueprint_data.search_warning_threshold,
        'select_warning_turn': blueprint_data.select_warning_turn,
        'select_warning_threshold': blueprint_data.select_warning_threshold,
        'personas': personas,
        'prev_persona_count': previous_personas_count,
        'max_times_persona_use': num_max_persona_use,
        'locations': locations,
        'pick_persona_with_replacement': blueprint_data.use_personas_with_replacement,
        'search_server': blueprint_data.search_server,
        'num_passages_retrieved': blueprint_data.num_passages_retrieved,
        'soft_block_qname': blueprint_data.block_qualification,
        constants.ROLE_QUALIFICATION_NAME_KEY: blueprint_data.role_qualification,
    }


def get_onboarding_world_opt(config: DictConfig):
    """
    Generates onboarding world opt from Mephisto config.
    """
    blueprint_data = config.mephisto.blueprint
    return {
        'wizard_time_out': blueprint_data.wizard_time_out,
        'apprentice_time_out': blueprint_data.apprentice_time_out,
        'send_task_data': False,
        'is_onboarding': True,
        'search_server': blueprint_data.search_server,
        'num_passages_retrieved': blueprint_data.num_passages_retrieved,
        'onboarding_qualification': blueprint_data.onboarding_qualification,
        constants.ROLE_QUALIFICATION_NAME_KEY: blueprint_data.role_qualification,
    }


def get_worker_eval_function(role_qname: str, onboarding_qname: str):
    """
    Returns the callback function that is used for checking worker qualification.

    Check `worker_can_do_unit` of `SharedTaskState` in Mephisto.
    """

    def worker_eval_function(worker, unit):
        """
        Checks the worker qualification for the task, based on their existing records.
        """
        worker_qualification = worker.get_granted_qualification(role_qname)
        if not worker_qualification:
            # has not done any onboarding training yet
            logging.debug('Worker does not have any qualifications (new worker).')
            return True

        qualification_status = worker_qualification.value
        logging.debug(f'Worker role qualification is {qualification_status}')
        if qualification_status in (
            constants.WIZARD_IN_TRAINING,
            constants.APPRENTICE_IN_TRAINING,
        ):
            # The agent had started the onboarding training but was not finished
            onboarding_qual = worker.get_granted_qualification(onboarding_qname)
            return not onboarding_qual or not onboarding_qual.value

        # The agent has successfully finished the onboarding training
        if unit.unit_index == 0:
            return qualification_status == constants.WIZARD
        else:
            return qualification_status == constants.APPRENTICE

    return worker_eval_function


def check_role_training_qualification(
    db: LocalMephistoDB, qname: str, requester_name: str
):
    """
    Initializes the qualification name in DB, if it does not exist.
    """

    logging.info(f'Checking for "{qname}"" qualification.')
    if not db.find_qualifications(qname):
        logging.info('Creating the qualification.')
        db.make_qualification(qname)
        reqs = db.find_requesters(requester_name=requester_name, provider_type='mturk')
        requester = reqs[-1]
        requester._create_new_mturk_qualification(qname)
    else:
        logging.info('Qualification exists.')


def update_persona_use_counts_file(
    fptah: str, counts: Dict[str, int], sorted_order=True
):
    """
    Writes the persona use counts to file.

    This is to keep track of use counts for the next time that the task was restarted.
    See `load_previously_used_personas_counts` function above.
    """
    logging.info(f'Writting new persona counts to {fptah}')
    items = counts.items()
    if sorted_order:
        items = sorted(items, key=lambda x: x[1], reverse=True)
    saved_count = 0
    with open(fptah, 'w') as fo:
        for p, c in items:
            if c > 0:
                saved_count += 1
                fo.write(f'{p} ; {c}\n')
    logging.info(f'Saved {saved_count} recent persona counts successfully.')


def add_banned_words_frontend_conf(task_state, fpath: str = None):
    """
    Adds the list of banned words to the task config to be used later in the frontend.

    It reads the text file specified in fpath to populate a list banned words. Then adds
    this list to Mephisto `task_config` to make it accessible for the front-end app. The
    file specified by `fpath` is a plain text file where each line contains a single
    banned word/phrase.
    """
    banned_words = []
    if fpath and os.path.exists(fpath):
        with open(fpath, 'r') as fin:
            banned_words.extend([w.strip().lower() for w in fin if w.strip()])

    task_state.task_config['bannedWords'] = banned_words


@hydra.main(config_name='scriptconfig')
def main(cfg: DictConfig) -> None:
    db, cfg = load_db_and_process_config(cfg)
    world_opt = get_world_opt(cfg)
    onboarding_world_opt = get_onboarding_world_opt(cfg)
    shared_state = SharedParlAITaskState(
        world_opt=world_opt, onboarding_world_opt=onboarding_world_opt
    )

    check_role_training_qualification(
        db=db,
        qname=world_opt[constants.ROLE_QUALIFICATION_NAME_KEY],
        requester_name=cfg.mephisto.provider.requester_name,
    )

    shared_state.task_config['minTurns'] = world_opt['min_turns']
    shared_state.task_config['onboardingPersona'] = constants.ONBOARDING_PERSONA
    shared_state.worker_can_do_unit = get_worker_eval_function(
        world_opt[constants.ROLE_QUALIFICATION_NAME_KEY],
        onboarding_world_opt['onboarding_qualification'],
    )

    banned_words_fpath = cfg.mephisto.blueprint.banned_words_file
    add_banned_words_frontend_conf(shared_state, banned_words_fpath)

    operator = Operator(db)
    operator.validate_and_run_config(cfg.mephisto, shared_state)
    operator.wait_for_runs_then_shutdown(skip_input=True, log_rate=300)
    update_persona_use_counts_file(
        cfg.mephisto.blueprint.persona_counts_file, world_opt['prev_persona_count']
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import argparse
import os
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List

import pandas as pd

from parlai.core.opt import Opt
import parlai.utils.logging as logging

# Defining the class only if Mephisto is installed, since it relies on Mephisto
try:
    from mephisto.abstractions.databases.local_database import LocalMephistoDB
    from mephisto.data_model.unit import Unit
    from mephisto.tools.data_browser import DataBrowser as MephistoDataBrowser
except ImportError:
    pass


class AbstractResultsCompiler(ABC):
    """
    Abstract class for compiling results of crowdsourcing runs.

    Currently only provides utility attributes/methods for analyzing turn annotations.
    """

    @classmethod
    def setup_args(cls):
        parser = argparse.ArgumentParser('Compile crowdsourcing results')
        parser.add_argument(
            '--output-folder', type=str, help='Folder to save output files to'
        )
        parser.add_argument(
            '--results-format',
            type=str,
            choices=['csv', 'json'],
            default='csv',
            help='Output format for results data',
        )
        return parser

    def __init__(self, opt: Opt):
        self.output_folder = opt.get('output_folder')
        self.results_format = opt['results_format']

    def get_results_path_base(self) -> str:
        """
        Return the save path for the results file, not including the file extension.
        """
        now = datetime.now()
        return os.path.join(
            self.output_folder,
            f'{self.__class__.__name__}__{now.strftime("%Y%m%d_%H%M%S")}',
        )

    def is_unit_acceptable(self, unit_data: Dict[str, Any]) -> bool:
        """
        Helps filtering units that are compiled. Override for use.

        Returning False means that the unit data will be discarded.
        """
        if not unit_data:
            # Add your task-specific qualificaiton logic that justifies
            # discarding this unit, based on it data content.
            return False

        return True

    @abstractmethod
    def compile_results(self) -> pd.DataFrame:
        """
        Method for returning the final results dataframe.

        Each row of the dataframe consists of one utterance of one conversation.
        """

    def compile_and_save_results(self):
        """
        Compile results and save them.

        Results will be saved in the format given by --results-format.
        """
        result_df = self.compile_results()
        results_path_base = self.get_results_path_base()
        results_path = f'{results_path_base}.{self.results_format}'
        os.makedirs(self.output_folder, exist_ok=True)
        if self.results_format == 'csv':
            result_df.to_csv(results_path, index=False)
        elif self.results_format == 'json':
            result_df.reset_index().to_json(results_path)
            # Reset the index to make each row have a unique index value
        else:
            raise ValueError(
                f'Results save format of "{self.results_format}" currently unsupported!'
            )
        print(f'Wrote results file to {results_path}.')


class AbstractTurnAnnotationResultsCompiler(AbstractResultsCompiler):
    """
    Results compiler subclass to provide utility code for turn annotations.

    Currently incompatible with Mephisto's DataBrowser: all subclasses load results
    files directly from disk.
    TODO: make all subclasses compatible with DataBrowser
    """

    @classmethod
    def setup_args(cls):
        parser = super().setup_args()
        parser.add_argument(
            '--results-folders', type=str, help='Comma-separated list of result folders'
        )
        parser.add_argument(
            '--problem-buckets',
            type=str,
            help='Comma-separated list of buckets used for annotation. Set to an empty string to not analyze problem buckets.',
            default='bucket_0,bucket_1,bucket_2,bucket_3,bucket_4,none_all_good',
        )
        return parser

    def __init__(self, opt: Opt):

        super().__init__(opt)

        # Handle inputs
        if 'results_folders' in opt:
            self.results_folders = opt['results_folders'].split(',')
        else:
            self.results_folders = None
        if opt['problem_buckets'].lower() not in ['', 'none']:
            self.use_problem_buckets = True
            self.problem_buckets = opt['problem_buckets'].split(',')
        else:
            self.use_problem_buckets = False
            self.problem_buckets = []


class AbstractDataBrowserResultsCompiler(AbstractResultsCompiler):
    """
    Provides interface for using Mephisto's DataBrowser, DB, and their methods.

    Uses Mephisto's DataBrowser to retrieve the work units and their data.
    """

    @classmethod
    def setup_args(cls):
        parser = super().setup_args()
        parser.add_argument(
            '--task-name', type=str, help='Name of the Mephisto task to open'
        )
        return parser

    def __init__(self, opt: Opt):
        super().__init__(opt)
        self.task_name = opt['task_name']
        self._mephisto_db = None
        self._mephisto_data_browser = None

    def get_mephisto_data_browser(self) -> MephistoDataBrowser:
        if not self._mephisto_data_browser:
            db = self.get_mephisto_db()
            self._mephisto_data_browser = MephistoDataBrowser(db=db)
        return self._mephisto_data_browser

    def get_mephisto_db(self) -> LocalMephistoDB:
        if not self._mephisto_db:
            self._mephisto_db = LocalMephistoDB()
        return self._mephisto_db

    def get_worker_name(self, worker_id: str) -> str:
        """
        Gets the global (AWS) id of a worker from their Mephisto worker_id.
        """
        db = self.get_mephisto_db()
        return db.get_worker(worker_id)["worker_name"]

    def get_task_units(self, task_name: str) -> List[Unit]:
        """
        Retrieves the list of work units from the Mephisto task.
        """
        data_browser = self.get_mephisto_data_browser()
        return data_browser.get_units_for_task_name(task_name)

    def get_data_from_unit(self, unit: Unit) -> Dict[str, Any]:
        """
        Retrieves task data for a single unit.
        """
        try:
            data_browser = self.get_mephisto_data_browser()
            return data_browser.get_data_from_unit(unit)
        except (IndexError, AssertionError):
            logging.warning(
                f'Skipping unit {unit.db_id}. No message found for this unit.'
            )

    def get_units_data(self, task_units: List[Unit]) -> List[Dict[str, Any]]:
        """
        Retrieves task data for a list of Mephisto task units.
        """
        task_data = []
        for unit in task_units:
            unit_data = self.get_data_from_unit(unit)
            if unit_data and self.is_unit_acceptable(unit_data):
                task_data.append(unit_data)

        return task_data



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from abc import ABC, abstractmethod, abstractproperty
from collections import defaultdict
from typing import List

from parlai.core.dict import DictionaryAgent
from parlai.core.opt import Opt
from parlai.utils.io import PathManager


try:
    from transformers import GPT2Tokenizer, T5TokenizerFast
except ImportError:
    raise ImportError(
        "Need to install Hugging Face transformers repository. "
        "Try `pip install transformers`."
    )

SPECIAL_TOKENS = {"bos_token": "<bos>", "eos_token": "<eos>", "pad_token": "<pad>"}

NO_OP = "x"


class HuggingFaceDictionaryAgent(DictionaryAgent, ABC):
    """
    Use Hugging Face tokenizers.
    """

    def __init__(self, opt: Opt, shared=None):
        if not shared:
            self.hf_tokenizer = self.get_tokenizer(opt)
            self.tok2ind = self.hf_tokenizer.get_vocab()
            self.ind2tok = {v: k for k, v in self.tok2ind.items()}
        else:
            self.hf_tokenizer = shared['hf_tokenizer']
            self.tok2ind = shared['tok2ind']
            self.ind2tok = shared['ind2tok']

        self.freq = defaultdict(int)
        for tok in self.tok2ind:
            self.freq[tok] = 1
        self.minfreq = opt.get('dict_minfreq', DictionaryAgent.default_minfreq)

        self._unk_token_idx = self.hf_tokenizer.unk_token_id
        self.override_special_tokens(opt)

        self.lower = opt.get('dict_lower', DictionaryAgent.default_lower)
        self.tokenizer = 'hf'
        self.opt = opt
        self.max_length = (
            self.opt.get('text_truncate') or self.hf_tokenizer.model_max_length
        )

    @abstractmethod
    def get_tokenizer(self, opt):
        """
        Instantiate the HuggingFace tokenizer for your model.
        """
        pass

    @abstractmethod
    def override_special_tokens(opt):
        """
        Override the special tokens for your tokenizer.
        """
        pass

    @abstractproperty
    def add_special_tokens(self) -> bool:
        """
        Whether to add special tokens when tokenizing.
        """

    @abstractproperty
    def skip_decode_special_tokens(self) -> bool:
        """
        Whether to skip special tokens when converting tokens to text.
        """

    def share(self):
        shared = super().share()
        shared['hf_tokenizer'] = self.hf_tokenizer
        shared['ind2tok'] = self.ind2tok
        shared['tok2ind'] = self.tok2ind
        return shared

    def format_text(self, text: str) -> str:
        """
        Format text prior to encoding with tokenizer.
        """
        return text

    def txt2vec(self, text, vec_type=list):
        return self.hf_tokenizer.encode(
            self.format_text(text),
            add_special_tokens=self.add_special_tokens,
            max_length=self.max_length,
            pad_to_max_length=False,
            truncation='longest_first',
        )

    def vec2txt(self, vec, **kwargs):
        return self.hf_tokenizer.decode(
            vec, skip_special_tokens=self.skip_decode_special_tokens, **kwargs
        )

    def act(self):
        return {}


class Gpt2DictionaryAgent(HuggingFaceDictionaryAgent):
    def is_prebuilt(self):
        """
        Indicates whether the dictionary is fixed, and does not require building.
        """
        return True

    @property
    def add_special_tokens(self) -> bool:
        """
        Whether to add special tokens when tokenizing.
        """
        return True

    @property
    def skip_decode_special_tokens(self) -> bool:
        """
        Whether to skip special tokens when converting tokens to text.
        """
        return False

    def get_tokenizer(self, opt):
        """
        Instantiate tokenizer.
        """
        model_sz = opt["gpt2_size"]
        if model_sz == "small":
            model_key = "gpt2"
        elif model_sz == "distilgpt2":
            model_key = "distilgpt2"
        else:
            model_key = f"gpt2-{model_sz}"
        # check if datapath has the files that hugging face code looks for
        hf_dir = os.path.join(opt["datapath"], "hf", model_key)
        if all(
            PathManager.exists(os.path.join(hf_dir, file_name))
            for file_name in ["merges.txt", "vocab.json"]
        ):
            fle_key = PathManager.get_local_path(hf_dir, recursive=True)

        else:
            fle_key = model_key
        return GPT2Tokenizer.from_pretrained(fle_key)

    def add_additional_special_tokens(self, additional_special_tokens: List[str]):
        """
        Add additional special tokens to the dictionary.
        """
        self.additional_special_tokens = additional_special_tokens
        self.hf_tokenizer.add_special_tokens(
            {'additional_special_tokens': additional_special_tokens}
        )
        for tok in self.additional_special_tokens:
            self.add_token(tok)

    def _define_special_tokens(self, opt):
        if opt["add_special_tokens"]:
            # Add additional start/end/pad tokens
            self.hf_tokenizer.add_special_tokens(SPECIAL_TOKENS)
            self.start_token = SPECIAL_TOKENS["bos_token"]
            self.end_token = SPECIAL_TOKENS["eos_token"]
            self.null_token = SPECIAL_TOKENS["pad_token"]
        else:
            # Only special token is end of text
            self.start_token = NO_OP  # hack, we cut off the start token
            self.end_token = "<|endoftext|>"
            self.null_token = "<|endoftext|>"

    def override_special_tokens(self, opt):
        # define special tokens
        self._define_special_tokens(opt)
        # now override
        self.start_idx = self.hf_tokenizer.convert_tokens_to_ids([self.start_token])[0]
        self.end_idx = self.hf_tokenizer.convert_tokens_to_ids([self.end_token])[0]
        self.null_idx = self.hf_tokenizer.convert_tokens_to_ids([self.null_token])[0]
        # set tok2ind for special tokens
        self.tok2ind[self.end_token] = self.end_idx
        self.tok2ind[self.start_token] = self.start_idx
        self.tok2ind[self.null_token] = self.null_idx
        # set ind2tok for special tokens
        self.ind2tok[self.end_idx] = self.end_token
        self.ind2tok[self.start_idx] = self.start_token
        self.ind2tok[self.null_idx] = self.null_token


class DialoGPTDictionaryAgent(Gpt2DictionaryAgent):
    def get_tokenizer(self, opt):
        """
        Instantiate tokenizer.
        """
        model_sz = opt["gpt2_size"]
        fle_key = f"microsoft/DialoGPT-{model_sz}"
        return GPT2Tokenizer.from_pretrained(fle_key)


class T5DictionaryAgent(HuggingFaceDictionaryAgent):
    def get_tokenizer(self, opt):
        return T5TokenizerFast.from_pretrained(opt['t5_model_arch'], truncation=True)

    @property
    def add_special_tokens(self) -> bool:
        """
        Whether to add special tokens when tokenizing.
        """
        return True

    @property
    def skip_decode_special_tokens(self) -> bool:
        """
        Whether to add special tokens when tokenizing.
        """
        return True

    def override_special_tokens(self, opt):
        # now override
        self.start_token = self.hf_tokenizer.pad_token
        self.end_token = self.hf_tokenizer.eos_token
        self.null_token = self.hf_tokenizer.pad_token
        self.unk_token = self.hf_tokenizer.unk_token

        self._unk_token_idx = self.hf_tokenizer.unk_token_id

        self.start_idx = self[self.start_token]
        self.end_idx = self[self.end_token]
        self.null_idx = self[self.null_token]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher
from .build import build
from parlai.utils.data import DatatypeHelper

import copy
import os
import codecs


def _path(opt, *additions):
    return os.path.join(opt['datapath'], 'CornellMovie', *additions)


class DefaultTeacher(DialogTeacher):
    DOUBLE = False

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        self.fold = DatatypeHelper.fold(opt['datatype'])
        build(opt)
        opt['datafile'] = _path(opt, self.fold + '.txt')
        super().__init__(opt, shared)

    def setup_data(self, datafile):
        lines_file = _path(self.opt, 'cornell movie-dialogs corpus', 'movie_lines.txt')
        convo_file = _path(
            self.opt, 'cornell movie-dialogs corpus', 'movie_conversations.txt'
        )

        lines = {}

        codecs.register_error('strict', codecs.ignore_errors)
        with codecs.open(lines_file, 'r') as f:
            for line in f:
                l = line.split(' +++$+++ ')
                lines[l[0]] = ' '.join(l[4:]).strip('\n').replace('\t', ' ')

        cnt = 0
        with codecs.open(convo_file, 'r') as f:
            for cnt, line in enumerate(f, 1):
                l = line.split(' ')
                convo = ' '.join(l[6:]).strip('\n').strip('[').strip(']')
                c = convo.replace("'", '').replace(' ', '').split(',')

                texts = [lines[l] for l in c]

                if (cnt % 10 == 0) and self.fold != 'test':
                    continue
                elif (cnt % 10 == 1) and self.fold != 'valid':
                    continue
                elif (cnt % 10 > 1) and self.fold != 'train':
                    continue

                for i, (prompt, response) in enumerate(zip(texts[::2], texts[1::2])):
                    yield {'text': prompt, 'label': response}, i == 0

                if self.DOUBLE:
                    for i, (prompt, response) in enumerate(
                        zip(texts[1::2], texts[2::2])
                    ):
                        yield {'text': prompt, 'label': response}, i == 0


class DoubleTeacher(DefaultTeacher):
    """
    This version creates text-label pairs from the perspective of both speakers.
    """

    DOUBLE = True



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Print pretty colors.
"""


class color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'
    RAINBOW = [RED, YELLOW, GREEN, CYAN, BLUE, PURPLE]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Module code for BlenderBot2.
"""
import random
import time
import torch
import torch.nn
from typing import List, Tuple, Dict, Optional

from parlai.agents.fid.fid import FidModel, T5FidModel, concat_enc_outs
from parlai.agents.rag.args import RetrieverType
from parlai.agents.rag.rag import RagModel, T5RagModel
from parlai.agents.rag.dpr import DprQueryEncoder, DprDocumentEncoder
from parlai.agents.rag.retrievers import (
    RagRetriever,
    Document,
    BLANK_DOC,
    argsort_scores_and_docs,
    retriever_factory as rag_retriever_factory,
    RagRetrieverTokenizer,
    SearchQuerySearchEngineRetriever,
    SearchQueryFAISSIndexRetriever,
)
from parlai.core.dict import DictionaryAgent
from parlai.core.opt import Opt
import parlai.utils.logging as logging
from parlai.utils.torch import padded_tensor

from .sub_modules import (
    QueryGenerator,
    MemoryDecoder,
    clean_vec_with_dict,
    RetrievalType,
    KnowledgeAccessMethod,
)


def retriever_factory(
    opt: Opt, dictionary: DictionaryAgent, shared=None
) -> Optional[RagRetriever]:
    """
    Build retriever.

    Override to build special BB2 Search Retrievers, if necessary

    :param opt:
        ParlAI Opt
    :param dictionary:
        dictionary agent
    :param shared:
        shared objects.

    :return retriever:
        return a retriever for RAG.
    """
    if opt.get('converting'):
        return None
    retriever = RetrieverType(opt['rag_retriever_type'])
    if retriever is RetrieverType.SEARCH_ENGINE:
        return BB2SearchQuerySearchEngineRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.SEARCH_TERM_FAISS:
        return BB2SearchQueryFaissIndexRetriever(opt, dictionary, shared=shared)
    else:
        return rag_retriever_factory(opt, dictionary, shared=shared)


class BlenderBot2RagModel(RagModel):
    """
    BlenderBot 2 RAG Model.

    Employs both a regular retriever and a long-term memory.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, retriever_shared=None):
        # TODO: Get rid of this hack
        opt['converting'] = True
        super().__init__(opt, dictionary, retriever_shared)
        opt['converting'] = False
        self.opt = opt
        self.dummy_retriever = DummyRetriever(opt, dictionary)
        self.retriever = retriever_factory(opt, dictionary, shared=retriever_shared)
        assert self.retriever is not None
        query_encoder = (
            self.retriever.query_encoder
            if hasattr(self.retriever, 'query_encoder')
            and opt['share_search_and_memory_query_encoder']
            else None
        )
        self.long_term_memory = LongTermMemory(
            opt, dictionary, query_encoder
        )  # type: ignore
        self.query_generator = QueryGenerator(opt)
        self.memory_decoder = MemoryDecoder(opt)

        # attrs
        self.knowledge_access_method = KnowledgeAccessMethod(
            opt['knowledge_access_method']
        )
        self.search = RetrieverType(opt['rag_retriever_type']) in [
            RetrieverType.SEARCH_ENGINE,
            RetrieverType.SEARCH_TERM_FAISS,
        ]
        self.should_generate_query = (
            self.knowledge_access_method is KnowledgeAccessMethod.CLASSIFY
            or self.search
        ) and (
            self.knowledge_access_method
            not in [KnowledgeAccessMethod.MEMORY_ONLY, KnowledgeAccessMethod.NONE]
        )

    def has_query_generator(self) -> bool:
        """
        Return whether there's a query generator.

        Directly access the query generator's agents.
        """
        return bool(self.query_generator.agents)

    def has_memory_decoder(self) -> bool:
        """
        Return whether there is a memory decoder.

        Directly access the memory decoder's agents.
        """
        return bool(self.memory_decoder.agents)

    def tokenize_query_generator_input(self, input: str) -> List[int]:
        """
        Tokenize the input for the query generator.
        """
        assert self.has_query_generator()
        return self.query_generator.tokenize_input(input)

    def tokenize_memory_decoder_input(self, input: str) -> List[int]:
        """
        Tokenize input for memory decoder.
        """
        assert self.has_memory_decoder()
        return self.memory_decoder.tokenize_input(input)

    def tokenize_memory(self, input: str) -> List[int]:
        """
        Tokenize input for Memory Retriever.
        """
        return self.long_term_memory.tokenize_query(input)

    def get_retrieval_type(self) -> torch.LongTensor:
        """
        Return retrieval type for current batch.

        Accesses the query generator directly.
        """
        assert self.has_query_generator()
        return self.query_generator.retrieval_type

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        memory_vec: torch.LongTensor,
        num_memories: torch.LongTensor,
        query_generator_vec: torch.LongTensor,
        gold_doc_vec: torch.LongTensor,
        gold_doc_title_vec: torch.LongTensor,
        num_gold_docs: torch.LongTensor,
        memory_decoder_vec: torch.LongTensor,
        num_memory_decoder_vecs: torch.LongTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        """
        Override RagModel.encoder to pass along several other input vecs.

        :param input:
            2D [bsz, seqlen] input to the encoder
        :param input_lengths:
            1D [bsz] lengths of each input item
        :param query_vec:
            2D [bsz*n_turns, seqlen] input for the retriever
        :param input_turns_cnt:
            1D [bsz] number of dialogue turns for each input example
        # Begin New Params
        :param memory_vec:
            3D [bsz, num_mems, seqlen] set of memories to write for each batch item
        :param num_memories:
            1D [bsz] # of memories per batch item
        :param query_generator_vec:
            2D [bsz, seqlen] input to the query generator
        :param gold_doc_vec:
            3D [bsz, num_docs, seqlen] gold documents per batch item
        :param gold_doc_title_vec:
            3D [bsz, num_docs, seqlen] gold document titles per batch item
        :param num_gold_docs:
            1D [bsz] # of gold documents per batch item
        :param memory_decoder_vec:
            3D [bsz, num_lines, seqlen] text to convert to memories with memory decoder
        :param num_memory_decoder_vecs:
            1D [bsz] # of memory decoder vectors for each batch item
        """
        # Retrieve, get expanded input
        if all([tensor is not None for tensor in [input_lengths, query_vec]]):
            expanded_input, top_docs, top_doc_scores = self.retrieve_and_concat(
                input,
                input_lengths,
                query_generator_vec,
                query_vec,
                input_turns_cnt,
                memory_vec,
                num_memories,
                gold_doc_vec,
                gold_doc_title_vec,
                num_gold_docs,
                memory_decoder_vec,
                num_memory_decoder_vecs,
            )
        else:
            expanded_input = input
            top_docs = top_doc_scores = None

        # Run through seq2seq encoder
        tensor, mask = self.seq2seq_encoder(
            expanded_input, positions, segments
        )  # type: ignore

        return tensor, mask, input_turns_cnt, top_docs, top_doc_scores

    def get_retrieval_indices(
        self, ret_vec: torch.LongTensor, ret_type: RetrievalType
    ) -> torch.LongTensor:
        """
        Return the batch indices for the given retrieval type.

        This function is extremely overloaded to handle all `KnowledgeAccessMethod`s and
        all `RetrievalType`s.

        Basically, if BB2's Access Method is not CLASSIFY, we return all indices
        if the specified retrieval type matches the access method. Otherwise, we look
        at the retrieval_type vector to find the corresponding indices.

        :param ret_vec:
            the retrieval_type vector indicating the "classified" retrieval type
            for each batch item
        :param ret_type:
            the retrieval type being considered here.

        :return indices:
            return which batch indices will utilize the given RetrievalType
        """
        no_indices = torch.zeros(0).long()
        all_indices = torch.arange(ret_vec.size(0)).long()
        type_indices = ret_vec.eq(ret_type.value).nonzero().squeeze(1).long()
        assert isinstance(all_indices, torch.LongTensor)
        assert isinstance(no_indices, torch.LongTensor)
        assert isinstance(type_indices, torch.LongTensor)

        if self.knowledge_access_method is KnowledgeAccessMethod.NONE:
            if ret_type is RetrievalType.NONE:
                return all_indices
            else:
                return no_indices
        elif self.knowledge_access_method is KnowledgeAccessMethod.ALL:
            if ret_type is RetrievalType.NONE:
                return no_indices
            else:
                return all_indices
        elif self.knowledge_access_method is KnowledgeAccessMethod.SEARCH_ONLY:
            if ret_type is RetrievalType.SEARCH:
                return all_indices
            else:
                return no_indices
        elif self.knowledge_access_method is KnowledgeAccessMethod.MEMORY_ONLY:
            if ret_type is RetrievalType.MEMORY:
                return all_indices
            else:
                return no_indices
        else:
            assert self.knowledge_access_method is KnowledgeAccessMethod.CLASSIFY
            return type_indices

    def retrieve_and_concat(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_generator_vec: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        memory_vec: torch.LongTensor,
        num_memories: torch.LongTensor,
        gold_doc_vec: torch.LongTensor,
        gold_doc_title_vec: torch.LongTensor,
        num_gold_docs: torch.LongTensor,
        memory_decoder_vec: torch.LongTensor,
        num_memory_decoder_vecs: torch.LongTensor,
    ) -> Tuple[torch.LongTensor, List[List[Document]], torch.Tensor]:
        """
        Override RagModel.retrieve_and_concat to perform different retrieval, depending
        on the.
        """
        start = time.time()
        logging.debug(f'Begin encoder: {time.time() - start:.2f}')
        if input_turns_cnt is not None:
            if query_generator_vec is not None:
                query_generator_vec = query_generator_vec.repeat_interleave(
                    input_turns_cnt, dim=0
                )  # type: ignore
            if memory_vec is not None:
                memory_vec = memory_vec.repeat_interleave(
                    input_turns_cnt, dim=0
                )  # type: ignore
            if num_memories is not None:
                num_memories = num_memories.repeat_interleave(
                    input_turns_cnt, dim=0
                )  # type: ignore
        n_input = (
            input_turns_cnt.sum().item()
            if input_turns_cnt is not None
            else input.size(0)
        )
        # 0a. Classify retrieval type, if necessary
        generated_memories = [[] for _ in range(int(n_input))]
        if memory_decoder_vec is not None:
            generated_memories = self.memory_decoder.generate_memories(
                memory_decoder_vec, num_memory_decoder_vecs
            )
        if self.should_generate_query:
            assert self.has_query_generator()
            retrieval_type, search_queries = self.query_generator.classify_retrieval(
                query_generator_vec, num_memories, generated_memories
            )
            logging.debug(f'Classify Retrieval: {time.time() - start:.2f}')
        else:
            retrieval_type = torch.LongTensor(input.size(0))
            search_queries = None

        # 1. Retrieve
        top_docs: List[List[Document]] = [[] for _ in range(int(n_input))]
        doc_scores: List[List[torch.Tensor]] = [[] for _ in range(int(n_input))]

        # 1a. retrieve from faiss or search
        search_indices = self.get_retrieval_indices(
            retrieval_type, RetrievalType.SEARCH
        )
        if search_indices.numel() > 0:
            search_docs, search_doc_scores = self.perform_search(
                search_queries, query_vec, search_indices
            )
            logging.debug(f'Search Complete: {time.time() - start:.2f}')
            logging.debug(f'search: {search_docs}')
            if gold_doc_vec is not None:
                logging.debug(f'num gold docs: {num_gold_docs}')
            self._fill_docs_and_scores(
                top_docs,
                doc_scores,
                search_indices,
                search_docs,
                search_doc_scores,
                gold_doc_vec,
                gold_doc_title_vec,
                num_gold_docs,
            )

        # 1b. memory search
        memory_indices = self.get_retrieval_indices(
            retrieval_type, RetrievalType.MEMORY
        )
        if memory_indices.numel() > 0:
            memories, memory_scores = self.access_long_term_memory(
                query_vec,
                memory_indices,
                memory_vec,
                num_memories,
                memory_decoder_vec,
                generated_memories,
            )
            logging.debug(f'Memory Access Complete: {time.time() - start:.2f}')
            if memories is not None and memory_scores is not None:
                self._fill_docs_and_scores(
                    top_docs, doc_scores, memory_indices, memories, memory_scores
                )

        # 1c. no search
        no_search_indices = self.get_retrieval_indices(
            retrieval_type, RetrievalType.NONE
        )
        if no_search_indices.numel() > 0:
            dummy_docs, dummy_scores = self.dummy_retriever.retrieve(
                query_vec[no_search_indices]  # type: ignore
            )
            logging.debug('no search')
            self._fill_docs_and_scores(
                top_docs, doc_scores, no_search_indices, dummy_docs, dummy_scores
            )

        # 2. Expand the input
        if input_turns_cnt is not None:
            input = input.repeat_interleave(input_turns_cnt, dim=0)  # type: ignore
            input_lengths = input_lengths.repeat_interleave(
                input_turns_cnt, dim=0
            )  # type: ignore
        top_doc_scores = torch.stack(
            [torch.cat([s_i for s_i in scores_i]) for scores_i in doc_scores]
        )
        expanded_input = self.concat_docs_and_input(
            input, input_lengths, top_docs, top_doc_scores.size(1)
        )
        return expanded_input, top_docs, top_doc_scores

    def _fill_docs_and_scores(
        self,
        top_docs: List[List[Document]],
        doc_scores: List[List[torch.Tensor]],
        indices: torch.LongTensor,
        docs: List[List[Document]],
        scores: torch.Tensor,
        gold_docs: Optional[torch.LongTensor] = None,
        gold_doc_titles: Optional[torch.LongTensor] = None,
        num_gold: Optional[torch.LongTensor] = None,
    ) -> Tuple[List[List[Document]], List[List[torch.Tensor]]]:
        """
        Fill top docs and doc scores with retrieved documents for L batch indices.

        Documents either come from a retriever or a long-term memory.

        :param top_docs:
            bsz-length list of document sets
        :param doc_scores:
            bsz x n_docs tensor of doc scores
        :param indices:
            batch indices (of length L) to consider
        :param docs:
            L-size list of document sets
        :param scores:
            L-size list of document scores
        :param gold_docs:
            optional list of gold documents to insert into the document set
        :param gold_doc_titles:
            corresponding optional list of doc titles
        :param num_gold:
            how many gold documents are provided per batch item.

        :return top_docs, doc_scores:
            return top_docs and doc_scores with documents and scores filled in
        """
        device = scores.device
        for idx, i in enumerate(indices):
            if num_gold is not None and num_gold[i] > 0:
                assert gold_doc_titles is not None
                assert gold_docs is not None
                replace_inds = random.sample(range(len(docs[idx])), int(num_gold[i]))
                for g_idx, replace in enumerate(replace_inds):
                    # the gold docs are tokenized with the retriever tokenizer.
                    # give the docs the max score, for rag models.
                    docs[idx][replace] = Document(
                        title=self.dict.vec2txt(
                            clean_vec_with_dict(self.dict, gold_doc_titles[i][g_idx])
                        ),
                        text=self.dict.vec2txt(
                            clean_vec_with_dict(self.dict, gold_docs[i][g_idx])
                        ),
                        docid='',
                    )
                    scores[idx][replace] = scores[idx][0].detach()
            try:
                top_docs[i] += docs[idx]
                if self.fp16:
                    scores = scores.half()
                doc_scores[i].append(scores[idx])
            except IndexError:
                # Docs are not provided for this example for this retrieval type.
                # Therefore, we assert that we are in the 'all' mode.
                top_docs[i] += [BLANK_DOC] * self.opt['n_docs']
                doc_scores[i].append(torch.ones(self.opt['n_docs']).to(device))

        return top_docs, doc_scores

    #################################
    # Retrieve/Concat Sub Functions #
    #################################

    def perform_search(
        self,
        search_queries: Optional[List[str]],
        query_vec: torch.LongTensor,
        search_indices: torch.LongTensor,
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieve via retriever from global knowledge.

        :param search_queries:
            if searching, a list of search queries
        :param query_vec:
            query encoder input if retrieving from FAISS
        :param search_indices:
            batch indices to search

        :return docs, scores:
            return the documents and corresponding retrieval scores.
        """
        assert self.retriever is not None
        if self.search:
            assert search_queries
            assert isinstance(
                self.retriever, SearchQuerySearchEngineRetriever
            ) or isinstance(self.retriever, SearchQueryFAISSIndexRetriever)
            self.retriever.set_search_queries(search_queries)
        search_docs, search_doc_scores = self.retriever.retrieve(
            query_vec[search_indices]  # type: ignore
        )
        return search_docs, search_doc_scores

    def access_long_term_memory(
        self,
        query_vec: torch.LongTensor,
        memory_indices: torch.LongTensor,
        memory_vec: Optional[torch.LongTensor],
        num_memories: torch.LongTensor,
        memory_decoder_vec: Optional[torch.LongTensor],
        generated_memories: List[List[str]],
    ) -> Tuple[Optional[List[List[Document]]], Optional[torch.Tensor]]:
        """
        Access long term memory.

        :param query_vec:
            retrieval vector for the long-term memory
        :param memory_indices:
            indices to access memory slots
        :param memory_vec:
            extracted memories from the observation
        :param num_memories:
            bsz-length tensor corresponding to number of memories per batch item
        :param memory_decoder_vec:
            input to the memory decoder
        :param generated_memories:
            memories generated by the memory decoder

        :return memories, memory_scores:
            return memories and memory scores, if there are memories retrieved
        """
        start = time.time()
        memories = None
        memory_scores = None
        memory_dict = {}
        indices = memory_indices.tolist()

        if memory_vec is not None:
            memory_dict = {
                batch_id: memory_vec[batch_id, : num_memories[mem_id]]
                for batch_id, mem_id in enumerate(indices)
            }
        if memory_decoder_vec is not None:
            for batch_id in indices:
                new_mems_i = generated_memories[batch_id]
                if not new_mems_i:
                    continue
                tokenized = [
                    self.long_term_memory.tokenize_query(m)
                    for m in generated_memories[batch_id]
                ]
                if batch_id in memory_dict:
                    tokenized += memory_dict[batch_id].tolist()
                new_mems_i, _ = padded_tensor(
                    tokenized, pad_idx=self.dict[self.dict.null_token]  # type: ignore
                )
                memory_dict[batch_id] = new_mems_i.to(query_vec)
        if self.knowledge_access_method in [
            KnowledgeAccessMethod.ALL,
            KnowledgeAccessMethod.MEMORY_ONLY,
        ]:
            # Add dummy memories just in case we are retrieving from memories
            if memory_vec is not None:
                seqlen = memory_vec.size(-1)
            elif memory_decoder_vec is not None:
                seqlen = memory_decoder_vec.size(-1)
            else:
                seqlen = query_vec.size(-1)
            for batch_id in indices:
                if batch_id not in memory_dict:
                    memory_dict[batch_id] = torch.zeros(1, seqlen).to(query_vec)
        if memory_dict:
            # first make sure all memories are padded to the same length.
            max_length = max([m.size(-1) for m in memory_dict.values()])
            for batch_id in memory_dict:
                vec = memory_dict[batch_id]
                if vec.size(-1) < max_length:
                    memory_dict[batch_id] = torch.cat(
                        [
                            vec,
                            torch.zeros((*vec.shape[:-1], max_length - vec.size(-1)))
                            .fill_(self.dict[self.dict.null_token])
                            .to(vec),
                        ],
                        dim=1,
                    )
            self.long_term_memory.write_memory(memory_dict)  # type: ignore
            logging.debug(f'Write Memory Complete: {time.time() - start:.2f}')
        if self.long_term_memory.has_memory():
            memories, memory_scores = self.long_term_memory.retrieve(
                query_vec[memory_indices]  # type: ignore
            )
            logging.debug(f'Memory Retrieval Complete: {time.time() - start:.2f}')
            logging.debug(f'memories: {memories}')
            logging.verbose('Reading from Memory')

        return memories, memory_scores


class T5BlenderBot2RagModel(T5RagModel, BlenderBot2RagModel):
    pass


class DummyRetriever(RagRetriever):
    """
    Dummy Retriever returns blank documents, and equal scores.

    It is utilized to pad the document numbers for mis-matched batch inputs.
    """

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Simply construct a list of blank documents for each query.
        """

        documents = [[BLANK_DOC] * self.opt['n_docs']] * query.size(0)
        scores = torch.ones(query.size(0), self.opt['n_docs']).to(query.device)
        return documents, scores


class LongTermMemory(RagRetriever):
    """
    The LongTermMEmory writes document embeddings to a memory.

    Retrieval then scores all documents in the memory and returns the final results.
    """

    def __init__(
        self,
        opt: Opt,
        dictionary: DictionaryAgent,
        query_encoder: Optional[torch.nn.Module] = None,
        shared=None,
    ):
        super().__init__(opt, dictionary, shared)
        self.n_docs = opt['n_docs']
        if query_encoder is None:
            self.query_encoder = DprQueryEncoder(
                opt,
                dpr_model=opt['memory_reader_model'],
                pretrained_path=opt['dpr_model_file'],
            )
        else:
            self.query_encoder = query_encoder
        self.memory_encoder = DprDocumentEncoder(
            opt,
            dpr_model=opt['memory_writer_model'],
            pretrained_path=opt['memory_writer_model_file'],
        ).eval()
        self._tokenizer = RagRetrieverTokenizer(
            query_model=opt['query_model'],
            dictionary=dictionary,
            delimiter='\n',
            max_length=opt['memory_retriever_truncate']
            if opt['memory_retriever_truncate'] > 0
            else opt['rag_query_truncate'],
        )
        self.max_memories = opt.get('max_memories', 100)
        self.num_memory_slots = opt.get('batchsize', 1) * opt.get('rag_turn_n_turns', 1)
        self.memory_vec_dict: Dict[int, torch.LongTensor] = {  # type: ignore
            k: torch.zeros(self.max_memories, opt['max_doc_token_length']).to(
                torch.int64
            )
            for k in range(self.num_memory_slots)
        }
        self.memory_enc_dict: Dict[int, torch.Tensor] = {
            k: torch.zeros(self.max_memories, opt['retriever_embedding_size'])
            for k in range(self.num_memory_slots)
        }
        self.active_memory_slots: List[int] = []
        self.dict = dictionary

    def has_memory(self) -> bool:
        """
        Return whether there is memory.
        """
        return bool(self.active_memory_slots)

    def write_memory(self, mem_dict: Dict[int, torch.LongTensor]):
        """
        Write vectors to memory.

        Assume that we clear the memory as well.

        :param mem_dict:
            mapping from memory slot to 2D-tokenized memories
        """
        self.active_memory_slots = list(mem_dict.keys())
        with torch.no_grad():
            slot_num_mems = [m.size(0) for m in mem_dict.values()]
            logging.debug(f'Writing {slot_num_mems} memories')
            mem_vecs = torch.cat(list(mem_dict.values()), dim=0)
            mem_encs = self.memory_encoder(mem_vecs)
            offset = 0
            for mem_slot, num_mems in zip(mem_dict.keys(), slot_num_mems):
                self.memory_vec_dict[mem_slot] = mem_vecs[  # type: ignore
                    offset : offset + num_mems
                ]
                self.memory_enc_dict[mem_slot] = mem_encs[offset : offset + num_mems]
                offset += num_mems

    def score_memories(self, query_enc: torch.Tensor) -> List[torch.Tensor]:
        """
        Score memories.
        """
        scores = []
        for i in range(query_enc.size(0)):
            scores.append(
                (
                    query_enc[i : i + 1]
                    @ self.memory_enc_dict[self.active_memory_slots[i]].t()
                ).squeeze(0)
            )
        return scores

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieve and score.

        Encode

        :param query:
            query tokens

        :return (docs, scores):
            docs: list of (text, title) tuples for each batch example
            scores: doc scores
        """
        query_enc = self.query_encoder(query)
        scores = self.score_memories(query_enc)

        top_docs, top_doc_scores = [], []
        for i in range(query.size(0)):
            scores_i = scores[i]
            memories_i, scores_i = argsort_scores_and_docs(
                scores_i, self.memory_vec_dict[i], self.n_docs  # type: ignore
            )
            mem_docs = []
            for mem in memories_i:
                mem_doc = Document('', self._tokenizer.decode(mem), '')  # type: ignore
                mem_doc.TITLE_DELIM = self.opt['memory_doc_title_delimiter']
                mem_docs.append(mem_doc)

            if len(mem_docs) < self.n_docs:
                # add dummy docs
                num_blank = self.n_docs - len(mem_docs)
                mem_docs += [BLANK_DOC] * num_blank
                scores_i = torch.cat([scores_i, torch.zeros(num_blank).to(scores_i)])
            top_docs.append(mem_docs)
            top_doc_scores.append(scores_i)
            logging.debug(scores_i)

        return top_docs, torch.stack(top_doc_scores)


class BlenderBot2FidModelMixin:
    embedding_size: int
    pad_idx: int

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        memory_vec: torch.LongTensor,
        num_memories: torch.LongTensor,
        query_generator_vec: torch.LongTensor,
        gold_doc_vec: torch.LongTensor,
        gold_doc_title_vec: torch.LongTensor,
        num_gold_docs: torch.LongTensor,
        memory_decoder_vec: torch.LongTensor,
        num_memory_decoder_vecs: torch.LongTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        enc_out, mask, input_turns_cnt, top_docs, top_doc_scores = super().encoder(  # type: ignore
            input,
            input_lengths,
            query_vec,
            input_turns_cnt,
            memory_vec,
            num_memories,
            query_generator_vec,
            gold_doc_vec,
            gold_doc_title_vec,
            num_gold_docs,
            memory_decoder_vec,
            num_memory_decoder_vecs,
            positions,
            segments,
        )  # type: ignore

        if input_turns_cnt is not None:
            # Input Turns is a tensor of dim [bsz]
            input = input.repeat_interleave(input_turns_cnt, dim=0)  # type: ignore

        new_out, new_mask = concat_enc_outs(
            input, enc_out, mask, self.embedding_size, self.pad_idx
        )

        return new_out, new_mask, input_turns_cnt, top_docs, top_doc_scores


class BlenderBot2FidModel(BlenderBot2FidModelMixin, BlenderBot2RagModel, FidModel):
    pass


class T5BlenderBot2FidModel(
    BlenderBot2FidModelMixin, T5FidModel, T5BlenderBot2RagModel
):
    pass


class BB2SearchRetrieverMixin:
    """
    Mixin for BB2 Search Modules.
    """

    def set_search_queries(self, queries: List[str]):
        self.search_queries = queries

    def init_search_query_generator(self, opt):
        pass

    def generate_search_query(self, query: torch.LongTensor) -> List[str]:
        return self.search_queries


class BB2SearchQuerySearchEngineRetriever(
    BB2SearchRetrieverMixin, SearchQuerySearchEngineRetriever
):
    """
    Override Search Engine Retriever to accommodate SQ Generator from BB2 Setup.
    """


class BB2SearchQueryFaissIndexRetriever(
    BB2SearchRetrieverMixin, SearchQueryFAISSIndexRetriever
):
    """
    Override Search Engine Retriever to accommodate SQ Generator from BB2 Setup.
    """
